import logging
from typing import List, Dict
import time
import datetime
from confluent_kafka import Consumer, KafkaError, KafkaException, TopicPartition
from csle_ryu.dao.avg_port_statistic import AvgPortStatistic
from csle_ryu.dao.avg_flow_statistic import AvgFlowStatistic
from csle_ryu.dao.agg_flow_statistic import AggFlowStatistic
from csle_ryu.dao.flow_statistic import FlowStatistic
from csle_ryu.dao.port_statistic import PortStatistic
import csle_collector.constants.constants as collector_constants
import csle_common.constants.constants as constants
from csle_collector.snort_ids_manager.dao.snort_ids_alert_counters import SnortIdsAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_rule_counters import SnortIdsRuleCounters
from csle_collector.snort_ids_manager.dao.snort_ids_ip_alert_counters import SnortIdsIPAlertCounters
from csle_collector.ossec_ids_manager.dao.ossec_ids_alert_counters import OSSECIdsAlertCounters
from csle_collector.client_manager.client_population_metrics import ClientPopulationMetrics
from csle_collector.docker_stats_manager.dao.docker_stats import DockerStats
from csle_collector.host_manager.dao.host_metrics import HostMetrics
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_config.emulation_metrics_time_series import EmulationMetricsTimeSeries


class ReadEmulationStatisticsUtil:
    """
    Utility class for reading emulation statistics
    """

    @staticmethod
    def read_all(emulation_env_config: EmulationEnvConfig, logger: logging.Logger, time_window_minutes: int = 100) \
            -> EmulationMetricsTimeSeries:
        """
        Reads all time series data from the kafka log

        :param emulation_env_config: the configuration of the emulation environment
        :param time_window_minutes : the length of the time window in minutes to consume (now-time_windows_minutes, now)
        :return: the collected time series data
        """
        client_metrics: List[ClientPopulationMetrics] = []
        agg_docker_stats: List[DockerStats] = []
        docker_host_stats: Dict[str, List[DockerStats]] = {}
        host_metrics: Dict[str, List[HostMetrics]] = {}
        aggregated_host_metrics: List[HostMetrics] = []
        defender_actions: List[EmulationDefenderAction] = []
        attacker_actions: List[EmulationAttackerAction] = []
        snort_ids_metrics: List[SnortIdsAlertCounters] = []
        snort_ids_ip_metrics: Dict[str, List[SnortIdsIPAlertCounters]] = {}
        snort_ids_rule_metrics: List[SnortIdsRuleCounters] = []
        snort_alert_metrics_per_ids: Dict[str, List[SnortIdsAlertCounters]] = {}
        snort_rule_metrics_per_ids: Dict[str, List[SnortIdsRuleCounters]] = {}
        total_snort_metrics: List[SnortIdsAlertCounters] = []
        total_snort_rule_metrics: List[SnortIdsRuleCounters] = []
        total_host_metrics: List[HostMetrics] = []
        ossec_host_ids_metrics: Dict[str, List[OSSECIdsAlertCounters]] = {}
        total_ossec_metrics: List[OSSECIdsAlertCounters] = []
        aggregated_ossec_metrics: List[OSSECIdsAlertCounters] = []
        openflow_flow_stats: List[FlowStatistic] = []
        openflow_port_stats: List[PortStatistic] = []
        avg_openflow_flow_stats: List[AvgFlowStatistic] = []
        avg_openflow_port_stats: List[AvgPortStatistic] = []
        openflow_flow_metrics_per_switch: Dict[str, List[FlowStatistic]] = {}
        openflow_port_metrics_per_switch: Dict[str, List[PortStatistic]] = {}
        openflow_flow_avg_metrics_per_switch: Dict[str, List[AvgFlowStatistic]] = {}
        openflow_port_avg_metrics_per_switch: Dict[str, List[AvgPortStatistic]] = {}
        agg_openflow_flow_stats: List[AggFlowStatistic] = []
        agg_openflow_flow_metrics_per_switch: Dict[str, List[AggFlowStatistic]] = {}

        num_ossec_containers = len(list(filter(lambda x: x.name in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES,
                                               emulation_env_config.containers_config.containers)))

        num_snort_containers = len(list(filter(lambda x: x.name in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES,
                                               emulation_env_config.containers_config.containers)))

        for c in emulation_env_config.containers_config.containers:
            docker_host_stats[c.get_full_name()] = []
            host_metrics[c.get_full_name()] = []
            ossec_host_ids_metrics[c.get_full_name()] = []
            snort_ids_ip_metrics[c.get_full_name()] = []
            for ids_image in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                if ids_image in c.name:
                    snort_alert_metrics_per_ids[c.get_full_name()] = []
                    snort_rule_metrics_per_ids[c.get_full_name()] = []

        host_metrics[emulation_env_config.kafka_config.container.get_full_name()] = []
        docker_host_stats[emulation_env_config.kafka_config.container.get_full_name()] = []
        ossec_host_ids_metrics[emulation_env_config.kafka_config.container.get_full_name()] = []
        snort_ids_ip_metrics[emulation_env_config.kafka_config.container.get_full_name()] = []
        host_metrics[emulation_env_config.elk_config.container.get_full_name()] = []
        docker_host_stats[emulation_env_config.elk_config.container.get_full_name()] = []
        ossec_host_ids_metrics[emulation_env_config.elk_config.container.get_full_name()] = []
        snort_ids_ip_metrics[emulation_env_config.elk_config.container.get_full_name()] = []
        if emulation_env_config.sdn_controller_config is not None:
            host_metrics[emulation_env_config.sdn_controller_config.container.get_full_name()] = []
            docker_host_stats[emulation_env_config.sdn_controller_config.container.get_full_name()] = []
            ossec_host_ids_metrics[emulation_env_config.sdn_controller_config.container.get_full_name()] = []
            snort_ids_ip_metrics[emulation_env_config.sdn_controller_config.container.get_full_name()] = []

        topic_names = [collector_constants.KAFKA_CONFIG.ATTACKER_ACTIONS_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.DOCKER_HOST_STATS_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.DEFENDER_ACTIONS_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.DOCKER_STATS_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.SNORT_IDS_LOG_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.HOST_METRICS_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.CLIENT_POPULATION_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.OSSEC_IDS_LOG_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.OPENFLOW_FLOW_STATS_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.OPENFLOW_PORT_STATS_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.AVERAGE_OPENFLOW_FLOW_STATS_PER_SWITCH_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.AVERAGE_OPENFLOW_PORT_STATS_PER_SWITCH_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.OPENFLOW_AGG_FLOW_STATS_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.SNORT_IDS_IP_LOG_TOPIC_NAME,
                       collector_constants.KAFKA_CONFIG.SNORT_IDS_RULE_LOG_TOPIC_NAME]
        logger.info(f"Reading time-series data for the last {time_window_minutes} minutes from topics: {topic_names}")
        start_consume_ts = time.time()
        kafka_conf = {
            collector_constants.KAFKA.BOOTSTRAP_SERVERS_PROPERTY:
                f"{emulation_env_config.kafka_config.container.docker_gw_bridge_ip}:"
                f"{emulation_env_config.kafka_config.kafka_port_external}",
            collector_constants.KAFKA.GROUP_ID_PROPERTY: f"emulation_data_consumer_thread_{start_consume_ts}",
            collector_constants.KAFKA.AUTO_OFFSET_RESET_PROPERTY: collector_constants.KAFKA.EARLIEST_OFFSET}
        consumer = Consumer(**kafka_conf)
        start_consume_ts = int(datetime.datetime.timestamp(datetime.datetime.now()
                                                           - datetime.timedelta(minutes=time_window_minutes)))
        start_consume_ts = int(start_consume_ts * 1e3)  # convert to ms
        topic_partitions = list(map(lambda x: TopicPartition(topic=x, partition=0, offset=start_consume_ts),
                                    topic_names))
        time_offsets_partitions = consumer.offsets_for_times(partitions=topic_partitions, timeout=20)

        def on_assign(consumer, partitions):
            """
            Callback called when a Kafka consumer is assigned to a partition
            """
            consumer.assign(time_offsets_partitions)

        consumer.subscribe(topic_names, on_assign=on_assign)
        done = False
        num_msg = 0
        host_metrics_counter = 0
        ossec_host_metrics_counter = 0
        snort_metrics_counter = 0
        snort_rule_metrics_counter = 0
        while not done:
            msg = consumer.poll(timeout=1.0)
            if msg is not None:
                num_msg += 1
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        logger.warning(f"reached end of partition: {msg.topic(), msg.partition(), msg.offset()}")
                    elif msg.error():
                        raise KafkaException(msg.error())
                else:
                    topic = msg.topic()
                    if topic == collector_constants.KAFKA_CONFIG.DOCKER_STATS_TOPIC_NAME:
                        agg_docker_stats.append(DockerStats.from_kafka_record(record=msg.value().decode()))
                    elif topic == collector_constants.KAFKA_CONFIG.HOST_METRICS_TOPIC_NAME:
                        metrics = HostMetrics.from_kafka_record(record=msg.value().decode())
                        c_1 = emulation_env_config.get_container_from_ip(metrics.ip)
                        if c_1 is not None:
                            host_metrics[c_1.get_full_name()].append(metrics)
                            host_metrics_counter += 1
                            total_host_metrics.append(metrics)
                    elif topic == collector_constants.KAFKA_CONFIG.OSSEC_IDS_LOG_TOPIC_NAME:
                        metrics = OSSECIdsAlertCounters.from_kafka_record(record=msg.value().decode())
                        c_1 = emulation_env_config.get_container_from_ip(metrics.ip)
                        if c_1 is not None:
                            ossec_host_ids_metrics[c_1.get_full_name()].append(metrics)
                            ossec_host_metrics_counter += 1
                            total_ossec_metrics.append(metrics)
                    elif topic == collector_constants.KAFKA_CONFIG.ATTACKER_ACTIONS_TOPIC_NAME:
                        attacker_actions.append(EmulationAttackerAction.from_kafka_record(record=msg.value().decode()))
                    elif topic == collector_constants.KAFKA_CONFIG.DEFENDER_ACTIONS_TOPIC_NAME:
                        defender_actions.append(EmulationDefenderAction.from_kafka_record(record=msg.value().decode()))
                    elif topic == collector_constants.KAFKA_CONFIG.DOCKER_HOST_STATS_TOPIC_NAME:
                        stats = DockerStats.from_kafka_record(record=msg.value().decode())
                        c_1 = emulation_env_config.get_container_from_ip(stats.ip)
                        if c_1 is not None:
                            docker_host_stats[c_1.get_full_name()].append(stats)
                    elif topic == collector_constants.KAFKA_CONFIG.SNORT_IDS_LOG_TOPIC_NAME:
                        metrics = SnortIdsAlertCounters.from_kafka_record(record=msg.value().decode())
                        c_1 = emulation_env_config.get_container_from_ip(metrics.ip)
                        if c_1 is not None:
                            snort_alert_metrics_per_ids[c_1.get_full_name()].append(metrics)
                            snort_metrics_counter += 1
                            total_snort_metrics.append(metrics)
                    elif topic == collector_constants.KAFKA_CONFIG.SNORT_IDS_RULE_LOG_TOPIC_NAME:
                        metrics = SnortIdsRuleCounters.from_kafka_record(record=msg.value().decode())
                        c_1 = emulation_env_config.get_container_from_ip(metrics.ip)
                        if c_1 is not None:
                            snort_rule_metrics_per_ids[c_1.get_full_name()].append(metrics)
                            snort_rule_metrics_counter += 1
                            total_snort_rule_metrics.append(metrics)
                    elif topic == collector_constants.KAFKA_CONFIG.CLIENT_POPULATION_TOPIC_NAME:
                        client_metrics.append(ClientPopulationMetrics.from_kafka_record(record=msg.value().decode()))
                    elif topic == collector_constants.KAFKA_CONFIG.OPENFLOW_FLOW_STATS_TOPIC_NAME:
                        flow_metrics_record = FlowStatistic.from_kafka_record(record=msg.value().decode())
                        openflow_flow_stats.append(flow_metrics_record)
                        if str(flow_metrics_record.datapath_id) not in openflow_flow_metrics_per_switch:
                            openflow_flow_metrics_per_switch[str(flow_metrics_record.datapath_id)] = []
                        openflow_flow_metrics_per_switch[str(flow_metrics_record.datapath_id)].append(
                            flow_metrics_record)
                    elif topic == collector_constants.KAFKA_CONFIG.OPENFLOW_PORT_STATS_TOPIC_NAME:
                        port_metrics_record = PortStatistic.from_kafka_record(record=msg.value().decode())
                        openflow_port_stats.append(port_metrics_record)
                        if str(port_metrics_record.datapath_id) not in openflow_port_metrics_per_switch:
                            openflow_port_metrics_per_switch[str(port_metrics_record.datapath_id)] = []
                        openflow_port_metrics_per_switch[str(port_metrics_record.datapath_id)].append(
                            port_metrics_record)
                    elif topic == collector_constants.KAFKA_CONFIG.AVERAGE_OPENFLOW_FLOW_STATS_PER_SWITCH_TOPIC_NAME:
                        avg_flow_statistics_record = AvgFlowStatistic.from_kafka_record(record=msg.value().decode())
                        avg_openflow_flow_stats.append(avg_flow_statistics_record)
                        if str(avg_flow_statistics_record.datapath_id) not in openflow_flow_avg_metrics_per_switch:
                            openflow_flow_avg_metrics_per_switch[str(avg_flow_statistics_record.datapath_id)] = []
                        openflow_flow_avg_metrics_per_switch[str(avg_flow_statistics_record.datapath_id)].append(
                            avg_flow_statistics_record)
                    elif topic == collector_constants.KAFKA_CONFIG.AVERAGE_OPENFLOW_PORT_STATS_PER_SWITCH_TOPIC_NAME:
                        avg_port_statistics_record = AvgPortStatistic.from_kafka_record(record=msg.value().decode())
                        avg_openflow_port_stats.append(avg_port_statistics_record)
                        if str(avg_port_statistics_record.datapath_id) not in openflow_port_avg_metrics_per_switch:
                            openflow_port_avg_metrics_per_switch[str(avg_port_statistics_record.datapath_id)] = []
                        openflow_port_avg_metrics_per_switch[str(avg_port_statistics_record.datapath_id)].append(
                            avg_port_statistics_record)
                    elif topic == collector_constants.KAFKA_CONFIG.OPENFLOW_AGG_FLOW_STATS_TOPIC_NAME:
                        agg_flow_statistics_record = AggFlowStatistic.from_kafka_record(record=msg.value().decode())
                        agg_openflow_flow_stats.append(agg_flow_statistics_record)
                        if str(agg_flow_statistics_record.datapath_id) not in agg_openflow_flow_metrics_per_switch:
                            agg_openflow_flow_metrics_per_switch[str(agg_flow_statistics_record.datapath_id)] = []
                        agg_openflow_flow_metrics_per_switch[str(agg_flow_statistics_record.datapath_id)].append(
                            agg_flow_statistics_record)
                    elif topic == collector_constants.KAFKA_CONFIG.SNORT_IDS_IP_LOG_TOPIC_NAME:
                        metrics = SnortIdsIPAlertCounters.from_kafka_record(record=msg.value().decode())
                        c_1 = emulation_env_config.get_container_from_ip(metrics.alert_ip)
                        if c_1 is not None:
                            snort_ids_ip_metrics[c_1.get_full_name()].append(metrics)
                    if host_metrics_counter >= len(emulation_env_config.containers_config.containers):
                        agg_host_metrics_dto = ReadEmulationStatisticsUtil.average_host_metrics(
                            host_metrics=total_host_metrics)
                        aggregated_host_metrics.append(agg_host_metrics_dto)
                        host_metrics_counter = 0
                        total_host_metrics = []
                    if ossec_host_metrics_counter >= num_ossec_containers:
                        agg_ossec_metrics_dto = ReadEmulationStatisticsUtil.average_ossec_metrics(
                            ossec_metrics=total_ossec_metrics)
                        aggregated_ossec_metrics.append(agg_ossec_metrics_dto)
                        ossec_host_metrics_counter = 0
                        total_ossec_metrics = []
                    if snort_metrics_counter >= num_snort_containers:
                        agg_snort_metrics_dto = ReadEmulationStatisticsUtil.average_snort_metrics(
                            snort_metrics=total_snort_metrics)
                        snort_ids_metrics.append(agg_snort_metrics_dto)
                        snort_metrics_counter = 0
                        total_snort_metrics = []
                    if snort_rule_metrics_counter >= num_snort_containers:
                        agg_snort_rule_metrics_dto = ReadEmulationStatisticsUtil.average_snort_rule_metrics(
                            snort_metrics=total_snort_rule_metrics)
                        snort_ids_rule_metrics.append(agg_snort_rule_metrics_dto)
                        snort_rule_metrics_counter = 0
                        total_snort_rule_metrics = []
            else:
                done = True
        consumer.close()
        dto = EmulationMetricsTimeSeries(
            client_metrics=client_metrics, aggregated_docker_stats=agg_docker_stats, host_metrics=host_metrics,
            docker_host_stats=docker_host_stats, agg_snort_ids_metrics=snort_ids_metrics,
            attacker_actions=attacker_actions,
            defender_actions=defender_actions, aggregated_host_metrics=aggregated_host_metrics,
            emulation_env_config=emulation_env_config, aggregated_ossec_host_alert_counters=aggregated_ossec_metrics,
            ossec_host_alert_counters=ossec_host_ids_metrics,
            avg_openflow_flow_stats=avg_openflow_flow_stats, avg_openflow_port_stats=avg_openflow_port_stats,
            openflow_port_stats=openflow_port_stats, openflow_flow_stats=openflow_flow_stats,
            openflow_flow_metrics_per_switch=openflow_flow_metrics_per_switch,
            openflow_port_metrics_per_switch=openflow_port_metrics_per_switch,
            openflow_flow_avg_metrics_per_switch=openflow_flow_avg_metrics_per_switch,
            openflow_port_avg_metrics_per_switch=openflow_port_avg_metrics_per_switch,
            agg_openflow_flow_stats=agg_openflow_flow_stats,
            agg_openflow_flow_metrics_per_switch=agg_openflow_flow_metrics_per_switch,
            agg_snort_ids_rule_metrics=snort_ids_rule_metrics, snort_ids_ip_metrics=snort_ids_ip_metrics,
            snort_rule_metrics_per_ids=snort_rule_metrics_per_ids,
            snort_alert_metrics_per_ids=snort_alert_metrics_per_ids
        )
        return dto

    @staticmethod
    def average_host_metrics(host_metrics: List[HostMetrics]) -> HostMetrics:
        """
        Computes the average metrics from a list of host metrics

        :param host_metrics: the list of host metrics to average
        :return: the computed averages
        """
        total_num_logged_in_users = 0
        total_num_failed_login_attempts = 0
        total_num_open_connections = 0
        total_num_login_events = 0
        total_num_processes = 0
        total_num_users = 0
        for metric in host_metrics:
            total_num_logged_in_users += metric.num_logged_in_users
            total_num_failed_login_attempts += metric.num_failed_login_attempts
            total_num_open_connections += metric.num_open_connections
            total_num_login_events += metric.num_login_events
            total_num_processes += metric.num_processes
            total_num_users += metric.num_users
        aggregated_host_metrics_dto = HostMetrics()
        aggregated_host_metrics_dto.num_logged_in_users = total_num_logged_in_users
        aggregated_host_metrics_dto.num_failed_login_attempts = total_num_failed_login_attempts
        aggregated_host_metrics_dto.num_open_connections = total_num_open_connections
        aggregated_host_metrics_dto.num_login_events = total_num_login_events
        aggregated_host_metrics_dto.num_processes = total_num_processes
        aggregated_host_metrics_dto.num_users = total_num_users
        return aggregated_host_metrics_dto

    @staticmethod
    def average_ossec_metrics(ossec_metrics: List[OSSECIdsAlertCounters]) -> OSSECIdsAlertCounters:
        """
        Computes the average metrics from a list of OSSEC metrics

        :param ossec_metrics: the list of OSSEC metrics
        :return: the computed averages
        """
        aggregated_ossec_ids_alert_counters = OSSECIdsAlertCounters()
        for alert_counters in ossec_metrics:
            aggregated_ossec_ids_alert_counters.add(alert_counters)
        return aggregated_ossec_ids_alert_counters

    @staticmethod
    def average_snort_metrics(snort_metrics: List[SnortIdsAlertCounters]) -> SnortIdsAlertCounters:
        """
        Computes the average metrics from a list of Snort metrics

        :param snort_metrics: the list of Snort metrics
        :return: the computed averages
        """
        aggregated_snort_ids_alert_counters = SnortIdsAlertCounters()
        for alert_counters in snort_metrics:
            aggregated_snort_ids_alert_counters.add(alert_counters)
        return aggregated_snort_ids_alert_counters

    @staticmethod
    def average_snort_rule_metrics(snort_metrics: List[SnortIdsRuleCounters]) -> SnortIdsRuleCounters:
        """
        Computes the average metrics from a list of Snort rule metrics

        :param snort_metrics: the list of Snort rule metrics
        :return: the computed averages
        """
        aggregated_snort_ids_rule_counters = SnortIdsRuleCounters()
        for alert_counters in snort_metrics:
            aggregated_snort_ids_rule_counters.add(alert_counters)
        return aggregated_snort_ids_rule_counters



from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
from csle_common.controllers.topology_controller import TopologyController


class TestTopologyControllerSuite:
    """
    Test suite for topology controller
    """

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("csle_common.util.emulation_util.EmulationUtil.disconnect_admin")
    def test_create_topology(self, mock_disconnect_admin, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test the utility function for connecting to a running emulation and creating the configuration

        :param mock_disconnect_admin: mock_disconnect_admin
        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :return: None
        """
        logger = MagicMock()
        # configure mock route and firewall
        mock_route = ("192.168.0.0/24", "192.168.1.1")
        mock_fw_config = MagicMock()
        mock_fw_config.physical_host_ip = "192.168.1.100"
        mock_fw_config.docker_gw_bridge_ip = "10.0.0.1"
        mock_fw_config.routes = [mock_route]
        mock_fw_config.ips_gw_default_policy_networks = []
        mock_fw_config.output_accept = []
        mock_fw_config.input_accept = []
        mock_fw_config.forward_accept = []
        mock_fw_config.output_drop = []
        mock_fw_config.input_drop = []
        mock_fw_config.forward_drop = []
        node_configs = [mock_fw_config]
        # configure mock topology and kafka
        topology_config = MagicMock()
        topology_config.node_configs = node_configs
        kafka_config = MagicMock()
        kafka_config.firewall_config = mock_fw_config
        # configure mock emulation environment
        emulation_env_config = MagicMock()
        emulation_env_config.topology_config = topology_config
        emulation_env_config.kafka_config = kafka_config
        emulation_env_config.sdn_controller_config = None
        emulation_env_config.get_connection.return_value = MagicMock()
        emulation_env_config.execution_id = "test_execution_id"
        emulation_env_config.level = "test_level"
        # set return values of execute_ssh_cmd
        mock_execute_ssh_cmd.return_value = (b"", b"", 0)
        TopologyController.create_topology(
            emulation_env_config=emulation_env_config, physical_server_ip="192.168.1.100", logger=logger)
        mock_connect_admin.assert_called()
        expected_cmd = f"{constants.COMMANDS.SUDO_ADD_ROUTE} 192.168.0.0/24 gw 192.168.1.1"
        mock_execute_ssh_cmd.assert_any_call(
            cmd=expected_cmd, conn=emulation_env_config.get_connection.return_value, wait_for_completion=True)
        mock_disconnect_admin.assert_called()
        logger.info.assert_any_call("Creating topology")
        logger.info.assert_any_call("Connecting to node:10.0.0.1")
        logger.info.assert_any_call(f"Adding route: {expected_cmd} to routing table of node: 10.0.0.1")



from typing import List, Dict, Any
from csle_common.dao.simulation_config.state import State
from csle_base.json_serializable import JSONSerializable


class StateSpaceConfig(JSONSerializable):
    """
    DTO representing the state space configuration of a simulation environment
    """

    def __init__(self, states: List[State]):
        """
        Initializes the DTO

        :param states: the list of states
        """
        self.states = states

    def states_ids(self) -> List[int]:
        """
        :return: list of state ids
        """
        return list(map(lambda x: x.id, self.states))

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "StateSpaceConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = StateSpaceConfig(states=list(map(lambda x: State.from_dict(x), d["states"])))
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d = {}
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"states: {list(map(lambda x: str(x), self.states))}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "StateSpaceConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return StateSpaceConfig.from_dict(json.loads(json_str))



import logging
from typing import List
import grpc
import time
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
import csle_common.constants.constants as constants
import csle_collector.constants.constants as csle_collector_constants
import csle_collector.five_g_core_manager.five_g_core_manager_pb2_grpc
import csle_collector.five_g_core_manager.five_g_core_manager_pb2
import csle_collector.five_g_core_manager.five_g_core_manager_util
import csle_collector.five_g_core_manager.query_five_g_core_manager
from csle_common.util.emulation_util import EmulationUtil
from csle_common.dao.emulation_config.five_g_core_managers_info import FiveGCoreManagersInfo


class FiveGCoreController:
    """
    Class controlling 5G cores running on nodes in the emulations, as well as 5G core managers
    """

    @staticmethod
    def start_five_g_core_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                   logger: logging.Logger) -> None:
        """
        Utility method for starting 5G core managers

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if ids_image in c.name:
                    FiveGCoreController.start_five_g_core_manager(emulation_env_config=emulation_env_config,
                                                                  ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_five_g_core_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility method for starting the 5G core manager on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Check if the manager is already running
        status = None
        not_running = False
        try:
            status = FiveGCoreController.get_five_g_core_status_by_ip_and_port(
                ip=ip, port=emulation_env_config.five_g_config.five_g_core_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = (f"mongo_running: {status.mongo_running}, mme_running: {status.mme_running}, "
                          f"sgwc_running: {status.sgwc_running}, smf_running: {status.smf_running}, "
                          f"amf_running: {status.amf_running}, sgwu_running: {status.sgwu_running}, "
                          f"upf_running: {status.upf_running}, hss_running: {status.hss_running}, "
                          f"pcrf_running: {status.pcrf_running}, nrf_running: {status.nrf_running}, "
                          f"scp_running: {status.scp_running}, sepp_running: {status.sepp_running}, "
                          f"ausf_running: {status.ausf_running}, udm_running: {status.udm_running}, "
                          f"pcf_running: {status.pcf_running}, nssf_running: {status.nssf_running}, "
                          f"bsf_running: {status.bsf_running}, udr_running: {status.udr_running}, "
                          f"webui_running: {status.webui_running}, ip: {status.ip}")
        if not_running:
            logger.info(f"Starting 5G core manager on node {ip}")

            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.FIVE_G_CORE_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))

            # Start the 5G core manager
            cmd = constants.COMMANDS.START_FIVE_G_CORE_MANAGER.format(
                emulation_env_config.five_g_config.five_g_core_manager_port,
                emulation_env_config.five_g_config.five_g_core_manager_log_dir,
                emulation_env_config.five_g_config.five_g_core_manager_log_file,
                emulation_env_config.five_g_config.five_g_core_manager_max_workers)
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            time.sleep(2)
        else:
            logger.info(f"5G core manager was already running on node {ip}. Status: {status_str}")

    @staticmethod
    def stop_five_g_core_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                  logger: logging.Logger) -> None:
        """
        Utility method for stopping 5G core managers

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the IP of the physical host
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if ids_image in c.name:
                    FiveGCoreController.stop_five_g_core_manager(emulation_env_config=emulation_env_config,
                                                                 ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_five_g_core_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility method for stopping a 5G core manager with a speicific IP

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

        logger.info(f"Stopping 5G Core manager on node {ip}")

        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.FIVE_G_CORE_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
        time.sleep(2)

    @staticmethod
    def get_five_g_core_manager_statuses(
            emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) \
            -> List[csle_collector.five_g_core_manager.five_g_core_manager_pb2.FiveGCoreStatusDTO]:
        """
        A method that sends a request to the 5G core manager on every container to get the status of the
        5G core

        :param emulation_env_config: the emulation config
        :param physical_server_ip: the IP of the physical server
        :param logger: the logger to use for logging
        :return: List of monitor thread statuses
        """
        statuses = []
        FiveGCoreController.start_five_g_core_managers(emulation_env_config=emulation_env_config,
                                                       physical_server_ip=physical_server_ip, logger=logger)

        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if ids_image in c.name:
                    status = FiveGCoreController.get_five_g_core_status_by_ip_and_port(
                        port=emulation_env_config.five_g_config.five_g_core_manager_port,
                        ip=c.docker_gw_bridge_ip)
                    statuses.append(status)
        return statuses

    @staticmethod
    def get_five_g_core_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPs of the 5G core managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        ips = []
        for c in emulation_env_config.containers_config.containers:
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if ids_image in c.name:
                    ips.append(c.docker_gw_bridge_ip)
        return ips

    @staticmethod
    def get_five_g_core_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the FiveGCoreManagers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of ports
        """
        ports = []
        for c in emulation_env_config.containers_config.containers:
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if ids_image in c.name:
                    ports.append(emulation_env_config.five_g_config.five_g_core_manager_port)
        return ports

    @staticmethod
    def get_five_g_core_status_by_ip_and_port(
            port: int, ip: str, timeout: int = csle_collector_constants.GRPC.TIMEOUT_SECONDS) \
            -> csle_collector.five_g_core_manager.five_g_core_manager_pb2.FiveGCoreStatusDTO:
        """
        A method that sends a request to the 5G core manager with a specific port and ip
        to get the status of the 5G core

        :param port: the port of the FiveGCoreManager
        :param ip: the ip of the FiveGCoreManager
        :param timeout: the timeout of the GRPC query
        :return: the status of the FiveGCoreManager
        """
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_core_manager.five_g_core_manager_pb2_grpc.FiveGCoreManagerStub(channel)
            status = \
                csle_collector.five_g_core_manager.query_five_g_core_manager.get_five_g_core_status(
                    stub=stub, timeout=timeout)
            return status

    @staticmethod
    def start_five_g_cores(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                           logger: logging.Logger) -> None:
        """
        Utility method for starting the 5G cores of a specific execution

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for container_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if container_image in c.name:
                    FiveGCoreController.start_five_g_core(emulation_env_config=emulation_env_config,
                                                          ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_five_g_core(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> csle_collector.five_g_core_manager.five_g_core_manager_pb2.FiveGCoreStatusDTO:
        """
        Utility method for starting the 5G core on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(f"Starting the 5G core on container with ip {ip} in execution {emulation_env_config.execution_id} "
                    f"of emulation: {emulation_env_config.name}")
        port = emulation_env_config.five_g_config.five_g_core_manager_port
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_core_manager.five_g_core_manager_pb2_grpc.FiveGCoreManagerStub(channel)
            status = csle_collector.five_g_core_manager.query_five_g_core_manager.start_five_g_core(stub=stub)
            return status

    @staticmethod
    def stop_five_g_cores(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                          logger: logging.Logger) -> None:
        """
        Utility method for stopping the 5G cores of a specific execution

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for container_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if container_image in c.name:
                    FiveGCoreController.stop_five_g_core(emulation_env_config=emulation_env_config,
                                                         ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_five_g_core(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> csle_collector.five_g_core_manager.five_g_core_manager_pb2.FiveGCoreStatusDTO:
        """
        Utility method for stopping the 5G core on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(f"Stopping the 5G core on container with ip {ip} in execution {emulation_env_config.execution_id} "
                    f"of emulation: {emulation_env_config.name}")
        port = emulation_env_config.five_g_config.five_g_core_manager_port
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_core_manager.five_g_core_manager_pb2_grpc.FiveGCoreManagerStub(channel)
            status = csle_collector.five_g_core_manager.query_five_g_core_manager.stop_five_g_core(stub=stub)
            return status

    @staticmethod
    def init_five_g_cores(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                          logger: logging.Logger) -> None:
        """
        Utility method for initializing the 5G cores of a specific execution

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for container_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if container_image in c.name:
                    FiveGCoreController.init_five_g_core(emulation_env_config=emulation_env_config,
                                                         ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def init_five_g_core(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> csle_collector.five_g_core_manager.five_g_core_manager_pb2.FiveGCoreStatusDTO:
        """
        Utility method for initializing the 5G core on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(
            f"Initializing the 5G core on container with ip {ip} in execution {emulation_env_config.execution_id} "
            f"of emulation: {emulation_env_config.name}")
        port = emulation_env_config.five_g_config.five_g_core_manager_port
        subscribers = list(map(lambda x: x.to_subscriber_dto(), emulation_env_config.five_g_config.subscribers))
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_core_manager.five_g_core_manager_pb2_grpc.FiveGCoreManagerStub(channel)
            status = csle_collector.five_g_core_manager.query_five_g_core_manager.init_five_g_core(
                stub=stub, subscribers=subscribers,
                core_backhaul_ip=emulation_env_config.five_g_config.core_backhaul_ip)
            return status

    @staticmethod
    def get_five_g_core_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                                      logger: logging.Logger, physical_server_ip: str) -> FiveGCoreManagersInfo:
        """
        Extracts the information of the 5G core managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param physical_server_ip: the IP of the physical server
        :param logger: the logger to use for logging
        :return: a DTO with the status of the 5G core managers
        """
        five_g_core_managers_ips = FiveGCoreController.get_five_g_core_managers_ips(
            emulation_env_config=emulation_env_config)
        five_g_core_managers_ports = FiveGCoreController.get_five_g_core_managers_ports(
            emulation_env_config=emulation_env_config)
        five_g_core_managers_statuses = []
        five_g_core_managers_running = []
        for ip in five_g_core_managers_ips:
            if ip not in active_ips or not EmulationUtil.physical_ip_match(
                    emulation_env_config=emulation_env_config, ip=ip, physical_host_ip=physical_server_ip):
                continue
            running = False
            status = None
            try:
                status = FiveGCoreController.get_five_g_core_status_by_ip_and_port(
                    port=emulation_env_config.five_g_config.five_g_core_manager_port, ip=ip)
                running = True
            except Exception as e:
                logger.debug(
                    f"Could not fetch 5G core manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
            if status is not None:
                five_g_core_managers_statuses.append(status)
            else:
                util = csle_collector.five_g_core_manager.five_g_core_manager_util.FiveGCoreManagerUtil
                five_g_core_managers_statuses.append(util.five_g_core_status_dto_empty())
            five_g_core_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        five_g_core_manager_info_dto = FiveGCoreManagersInfo(
            five_g_core_managers_running=five_g_core_managers_running, ips=five_g_core_managers_ips,
            ports=five_g_core_managers_ports, execution_id=execution_id, emulation_name=emulation_name,
            five_g_core_managers_statuses=five_g_core_managers_statuses)
        return five_g_core_manager_info_dto

    @staticmethod
    def start_five_g_core_monitor_threads(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                          logger: logging.Logger) -> None:
        """
        A method that sends a request to the 5G CoreManager on every container
        to start the Core manager and the monitor thread

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for container_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if container_image in c.name:
                    FiveGCoreController.start_five_g_core_monitor_thread(emulation_env_config=emulation_env_config,
                                                                         ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_five_g_core_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> None:
        """
        A method that sends a request to the FiveGCoreManager on a specific IP
        to start the 5G core monitor thread

        :param emulation_env_config: the emulation env config
        :param ip: IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        core_status_dto = FiveGCoreController.get_five_g_core_status_by_ip_and_port(
            ip=ip, port=emulation_env_config.five_g_config.five_g_core_manager_port)
        if not core_status_dto.monitor_running:
            logger.info(f"5G Core monitor thread is not running on {ip}, starting it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ip}:{emulation_env_config.five_g_config.five_g_core_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.five_g_core_manager.five_g_core_manager_pb2_grpc.FiveGCoreManagerStub(channel)
                csle_collector.five_g_core_manager.query_five_g_core_manager.start_five_g_core_monitor(
                    stub=stub, kafka_ip=emulation_env_config.kafka_config.container.get_ips()[0],
                    kafka_port=emulation_env_config.kafka_config.kafka_port,
                    time_step_len_seconds=emulation_env_config.kafka_config.time_step_len_seconds)

    @staticmethod
    def stop_five_g_core_monitor_threads(emulation_env_config: EmulationEnvConfig, logger: logging.Logger,
                                         physical_host_ip: str) -> None:
        """
        A method that sends a request to the 5G core on every container to stop the monitor threads

        :param emulation_env_config: the emulation env config
        :param physical_host_ip: the IP of the physical host
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_host_ip:
                continue
            for container_image in constants.CONTAINER_IMAGES.FIVE_G_CORE_IMAGES:
                if container_image in c.name:
                    FiveGCoreController.stop_five_g_core_monitor_thread(emulation_env_config=emulation_env_config,
                                                                        ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_five_g_core_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> None:
        """
        A method that sends a request to the 5G Core Manager on a specific container to stop the monitor thread

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Open a gRPC session
        with grpc.insecure_channel(f'{ip}:{emulation_env_config.five_g_config.five_g_core_manager_port}',
                                   options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_core_manager.five_g_core_manager_pb2_grpc.FiveGCoreManagerStub(channel)
            logger.info(f"Stopping the 5G core monitor thread on {ip}.")
            csle_collector.five_g_core_manager.query_five_g_core_manager.stop_five_g_core_monitor(stub=stub)



from typing import List, Dict, Any
from csle_common.dao.emulation_action_result.nmap_hop import NmapHop
from csle_base.json_serializable import JSONSerializable


class NmapTrace(JSONSerializable):
    """
    DTO Representing an NMAP Trace
    """

    def __init__(self, hops: List[NmapHop]):
        """
        Initializes the DTO

        :param hops: the list of hops in the trace
        """
        self.hops = hops

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapTrace":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapTrace(
            hops=list(map(lambda x: NmapHop.from_dict(x), d["hops"]))
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d = {}
        d["hops"] = list(map(lambda x: x.to_dict(), self.hops))
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "hops:{}".format(list(map(lambda x: str(x), self.hops)))

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapTrace":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapTrace.from_dict(json.loads(json_str))

    @staticmethod
    def schema() -> "NmapTrace":
        """
        :return: get the schema of the DTO
        """
        return NmapTrace(hops=[NmapHop.schema()])



from typing import List, Dict, Any
from csle_common.dao.simulation_config.value_type import ValueType
from csle_common.dao.simulation_config.observation import Observation
from csle_base.json_serializable import JSONSerializable


class ObservationSpaceConfig(JSONSerializable):
    """
    DTO representing the configuration of the observation space of a single player in a simulation environment
    """

    def __init__(self, observations: List[Observation], observation_type: ValueType, descr: str,
                 player_id: int, observation_component_name_to_index: Dict[str, int],
                 observation_id_to_observation_id_vector: Dict[int, List[int]],
                 observation_id_to_observation_vector: Dict[int, List[int]],
                 component_observations: Dict[str, List[Observation]]):
        """
        Initializes the DTO

        :param observations: the list of observations
        :param observation_type: the observation type
        :param descr: a description of the observation space
        :param player_id: the id of the player
        :param observation_component_name_to_index
        :param observation_id_to_observation_id_vector
        :param component_observations: mapping between observation component names and sub-observation spaces
        """
        self.observations = observations
        self.observation_type = observation_type
        self.descr = descr
        self.player_id = player_id
        self.observation_component_name_to_index = observation_component_name_to_index
        self.observation_id_to_observation_id_vector = observation_id_to_observation_id_vector
        self.observation_id_to_observation_id_vector_inv = dict(zip(
            list(map(lambda x: ",".join(list(map(lambda y: str(y), x))),
                     self.observation_id_to_observation_id_vector.values())),
            self.observation_id_to_observation_id_vector.keys()))
        self.observation_id_to_observation_vector = observation_id_to_observation_vector
        self.observation_id_to_observation_vector_inv = dict(
            zip(list(map(lambda x: ",".join(list(map(lambda y: str(y), x))),
                         self.observation_id_to_observation_id_vector.values())),
                self.observation_id_to_observation_vector.keys()))
        self.component_observations = component_observations

    def observation_ids(self) -> List[int]:
        """
        :return: a list of observation ids
        """
        return list(map(lambda x: x.id, self.observations))

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ObservationSpaceConfig":
        """
        Converts a dict representation to an instance
        :param d: the dict to convert
        :return: the created instance
        """
        component_observations = {}
        for k, v in d["component_observations"].items():
            component_observations[k] = list(map(lambda x: Observation.from_dict(x), v))

        obj = ObservationSpaceConfig(
            observations=list(map(lambda x: Observation.from_dict(x), d["observations"])),
            observation_type=d["observation_type"], descr=d["descr"],
            player_id=d["player_id"], observation_component_name_to_index=d["observation_component_name_to_index"],
            observation_id_to_observation_id_vector=d["observation_id_to_observation_id_vector"],
            observation_id_to_observation_vector=d["observation_id_to_observation_vector"],
            component_observations=component_observations
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["observations"] = list(map(lambda x: x.to_dict(), self.observations))
        d["observation_type"] = self.observation_type
        d["descr"] = self.descr
        d["player_id"] = self.player_id
        d["observation_component_name_to_index"] = self.observation_component_name_to_index
        d["observation_id_to_observation_vector"] = self.observation_id_to_observation_vector
        d["observation_id_to_observation_id_vector"] = self.observation_id_to_observation_id_vector
        d2 = {}
        for k, v in self.component_observations.items():
            d2[k] = list(map(lambda x: x.to_dict(), v))
        d["component_observations"] = d2
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"observations: {self.observations}, observation_type: {self.observation_type}, " \
               f"descr: {self.descr}, player_id: {self.player_id}, " \
               f"observation_component_name_to_index: {self.observation_component_name_to_index}, " \
               f"observation_id_to_observation_vector: {self.observation_id_to_observation_vector}," \
               f"component_observations: {self.component_observations}, " \
               f"observation_id_to_observation_vector: {self.observation_id_to_observation_vector}," \
               f"observation_id_to_observation_id_vector: {self.observation_id_to_observation_id_vector}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ObservationSpaceConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ObservationSpaceConfig.from_dict(json.loads(json_str))



from typing import Tuple, Any, Dict
import io
import docker
import logging
import pytest
import pytest_mock
from csle_common.controllers.container_controller import ContainerController
import csle_common.constants.constants as constants
from csle_collector.docker_stats_manager.docker_stats_manager_pb2 import DockerStatsMonitorDTO
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.emulation_execution import EmulationExecution
from csle_common.dao.emulation_config.containers_config import ContainersConfig
from csle_common.dao.emulation_config.docker_stats_manager_config import DockerStatsManagerConfig
from csle_common.dao.emulation_config.docker_stats_managers_info import DockerStatsManagersInfo
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.dao.emulation_config.config import Config
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.docker.docker_env_metadata import DockerEnvMetadata


class TestContainerControllerSuite:
    """
    Test suite for the container controller
    """

    @pytest.fixture
    def client_1(self, mocker: pytest_mock.MockFixture) -> docker.from_env:
        """
        Pytest fixture for mocking the docker.from_env() method
        
        :param mocker: the Pytest mocker object
        :return: the mocked function
        """

        class from_env:
            def __init__(self) -> None:
                pass

            class containers:

                @staticmethod
                def list(all=False):
                    class image:
                        def __init__(self) -> None:
                            self.short_id = '1'
                            self.tags = ['tags']

                    class element:

                        def __init__(self) -> None:
                            self.name = constants.CONTAINER_IMAGES.CSLE_PREFIX + 'null' + '-' + 'level'
                            self.name = self.name + constants.CSLE.NAME + '--1'
                            self.status = constants.DOCKER.CONTAINER_EXIT_STATUS
                            self.labels = {constants.DOCKER.CFG: 'config_path',
                                           constants.DOCKER.CONTAINER_CONFIG_DIR: 'dir_path',
                                           constants.DOCKER.EMULATION: "JDoeEmulation"}
                            self.id = '1'
                            self.short_id = '1'
                            self.image = image()

                        def stop(self):
                            return None

                        def remove(self):
                            return None

                        def start(self):
                            return None

                    return [element()]

            class networks:

                @staticmethod
                def list():
                    class element:
                        def __init__(self) -> None:
                            self.name = constants.CONTAINER_IMAGES.CSLE_PREFIX + 'null' + '-' + 'level'
                            self.name = self.name + constants.CSLE.NAME + '--1'

                        def remove(self, ):
                            return None

                    return [element()]

                @staticmethod
                def create(name: str, driver: str, ipam: docker.types.IPAMConfig, attachable: bool):
                    return None

            class images:

                @staticmethod
                def list():
                    class element:
                        def __init__(self) -> None:
                            self.attrs = {constants.DOCKER.REPO_TAGS: [constants.CSLE.NAME,
                                                                       constants.DOCKER.BASE_CONTAINER_TYPE,
                                                                       constants.OS.UBUNTU, constants.OS.KALI],
                                          constants.DOCKER.IMAGE_CREATED: "yes",
                                          constants.DOCKER.IMAGE_OS: "os",
                                          constants.DOCKER.IMAGE_ARCHITECTURE: "csle-architecture",
                                          constants.DOCKER.IMAGE_SIZE: 100}
                            self.short_id = '1'
                            self.tags = ['tags']
                            self.name = constants.CSLE.NAME

                    return [element()]

                @staticmethod
                def remove(image, force):
                    return None

        from_env_mocker = mocker.MagicMock(side_effect=from_env)
        return from_env_mocker

    @pytest.fixture
    def ipam_pool(self, mocker: pytest_mock.MockFixture) -> docker.types.IPAMPool:
        """
        Pytest fixture for mocking the docker.types.IPAMPool class
        
        :param mocker: the pytest mocker object
        :return: the mocked object
        """

        class IPAMPool():
            def __init__(self, subnet) -> None:
                pass

        ipampool_mocker = mocker.MagicMock(side_effect=IPAMPool)
        return ipampool_mocker

    @pytest.fixture
    def ipam_config(self, mocker: pytest_mock.MockFixture) -> docker.types.IPAMConfig:
        """
        Pytest fixture for mocking the docker.types.ImpamConfig class

        :param mocker: the pytest mocker object
        :return: the mocked object
        """

        class IPAMConfig():
            def __init__(self, pool_configs) -> None:
                pass

        ipamconfig_mocker = mocker.MagicMock(side_effect=IPAMConfig)
        return ipamconfig_mocker

    @pytest.fixture
    def client_2(self, mocker: pytest_mock.MockFixture):
        """
        Pytest fixture for mocking the docker.APIClient
        
        :param mocker: the pytest mocker object
        :return: the mocked object
        """

        class APIClient():
            def __init__(self, base_url) -> None:
                self.base_url = base_url

            def inspect_container(self, param: int) -> Dict[str, Any]:
                dict = {
                    constants.DOCKER.NETWORK_SETTINGS: {
                        constants.DOCKER.NETWORKS: {
                            'net_key': {
                                constants.DOCKER.IP_ADDRESS_INFO: "123.456.78.99", constants.DOCKER.NETWORK_ID_INFO: 1,
                                constants.DOCKER.GATEWAY_INFO: "null", constants.DOCKER.MAC_ADDRESS_INFO: "null",
                                constants.DOCKER.IP_PREFIX_LEN_INFO: 1}
                        }
                    },
                    constants.DOCKER.CREATED_INFO: "created_info",
                    constants.DOCKER.CONFIG: {
                        constants.DOCKER.HOSTNAME_INFO: "JDoeHost",
                        constants.DOCKER.IMAGE: "JDoeImage"}
                }
                return dict

        api_mocker = mocker.MagicMock(side_effect=APIClient)
        return api_mocker

    @pytest.fixture
    def file_opener(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the os.popen method

        :param mocker: the pytest mocker object
        :return: the mocked function
        """

        def popen(cmd: str):
            file = io.StringIO()
            file.write(f" test {constants.CSLE.CSLE_NETWORK_PREFIX}1 \n")
            file.write(f" test {constants.CSLE.CSLE_NETWORK_PREFIX}2\n")
            file.seek(0)
            return file

        popen_mocker = mocker.MagicMock(side_effect=popen)
        return popen_mocker

    @pytest.fixture
    def sub_popen(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the suboprocess.Popen class

        :param mocker: the pytest mocker object
        :return: the mocked version
        """

        class Popen():
            def __init__(self, cmd: str, stdout, shell: bool) -> None:
                pass

            def __iter__(self, ):
                pass

            def communicate(self) -> Tuple[str, None]:
                return (constants.COMMANDS.SEARCH_DOCKER_STATS_MANAGER + '37752', None)

        sub_popen_mocker = mocker.MagicMock(side_effect=Popen)
        return sub_popen_mocker

    @pytest.fixture
    def true_running_stats_manager(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the is_statsmanager_running method

        :param mocker: the pytest mocker object
        :return: True from the mocked function
        """

        def is_statsmanager_running() -> bool:
            return True

        is_statsmanager_running_mock = mocker.MagicMock(side_effect=is_statsmanager_running)
        return is_statsmanager_running_mock

    @pytest.fixture
    def false_running_stats_manager(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the is_statsmanager_running method

        :param mocker: the pytest mocker object
        :return: False from the mocked function
        """

        def is_statsmanager_running() -> bool:
            return False

        is_statsmanager_running_mock = mocker.MagicMock(side_effect=is_statsmanager_running)
        return is_statsmanager_running_mock

    @pytest.fixture
    def pid_file(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the read_pid_file in the ManagementSystemController

        :param mocker: the pytest mocker object
        :return: the mocked function
        """

        def read_pid_file(path: str) -> int:
            return 100

        pid_file_mocker = mocker.MagicMock(side_effect=read_pid_file)
        return pid_file_mocker

    @pytest.fixture
    def start_fixture(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the start_docker_stats_monitor in the csle_collector environment
        :param mocker: the pytest mocker object
        :return: None
        """

        def start_docker_stats_monitor(stub: str, emulation: EmulationEnvConfig,
                                       kafka_ip: str, stats_queue_maxsize: int,
                                       time_step_len_seconds: int,
                                       kafka_port: str, containers: NodeContainerConfig,
                                       execution_first_ip_octet: str) -> None:
            return None

        start_mocker = mocker.MagicMock(side_effect=start_docker_stats_monitor)
        return start_mocker

    @pytest.fixture
    def host_ip(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the get_host_ip method
        
        :param mocker: the pytest mocker object
        :return: the mocked function
        """

        def get_host_ip() -> str:
            return "123.456.78.99"

        host_ip_mocker = mocker.MagicMock(side_effect=get_host_ip)
        return host_ip_mocker

    @pytest.fixture
    def stop_fixture(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the stop_docker_stats_monitor in the csle_collector environment

        :param mocker: the pytest mocker object
        :return: None
        """

        def stop_docker_stats_monitor(stub: str, emulation: str, execution_first_ip_octet: str) -> None:
            return None

        stop_mocker = mocker.MagicMock(side_effect=stop_docker_stats_monitor)
        return stop_mocker

    @pytest.fixture
    def true_stop_fixture(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the stop_docker_statsmanager method

        :param mocker: the pytest mocker object
        :return: the mocked function
        """

        def stop_docker_statsmanager(logger: logging.Logger) -> bool:
            return True

        stop_fixture = mocker.MagicMock(side_effect=stop_docker_statsmanager)
        return stop_fixture

    @pytest.fixture
    def false_stop_fixture(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the stop_docker_statsmanager method

        :param mocker: the pytest mocker object
        :return: the mocked function
        """

        def stop_docker_statsmanager(logger: logging.Logger) -> bool:
            return False

        stop_fixture = mocker.MagicMock(side_effect=stop_docker_statsmanager)
        return stop_fixture

    @pytest.fixture
    def dsm_status(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the get_docker_stats_manager_status method in the csle_collector environment

        :param mocker: the pytest mocker object
        :return: the mocked function
        """

        def get_docker_stats_manager_status(stub) -> DockerStatsMonitorDTO:
            return DockerStatsMonitorDTO(num_monitors=10, emulations=['JDoeEmulation'], emulation_executions=[10])

        dsm_status_mocker = mocker.MagicMock(side_effect=get_docker_stats_manager_status)
        return dsm_status_mocker

    @pytest.fixture
    def stub(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the parameter stub fetched from csle_collector

        :param mocker: the pytest mocker object
        :return: the mocked object
        """

        class DockerStatsManagerStub():
            def __init__(self, channel) -> None:
                pass

        stub_mocker = mocker.MagicMock(side_effect=DockerStatsManagerStub)
        return stub_mocker

    @pytest.fixture
    def socket_fix(self, mocker: pytest_mock.MockFixture) -> Any:
        """
        Pytest fixture for mocking the imoprted socket module

        :param mocker: the pytest mocker object
        :return: the mocked object
        """

        class socket():
            def __init__(self, family: int = -1, type: int = -1, proto: int = -1, fileno: Any = None) -> None:
                self.family = family
                self.type = type
                self.proto = proto
                self.fileno = fileno

            def connect(self, a) -> None:
                return None

            def getsockname(self):
                return ["123.456.78.99"]

        socket_mocker = mocker.MagicMock(side_effect=socket)
        return socket_mocker

    def test_stop_all_running_containers(self, mocker: pytest_mock.MockFixture, client_1) -> None:
        """
        Tests the stop_all_running_containers method of the ContainerController

        :param mocker: the Pytest mock object
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)

        assert ContainerController.stop_all_running_containers() is None

    def test_stop_container(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture) -> None:
        """
        Tests the stop_container method od the ContainerController

        :param mocker: the Pytest mock object
        :param client_1: fixture for mocking the docker.from_env class
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        assert ContainerController.stop_container(constants.CONTAINER_IMAGES.CSLE_PREFIX +
                                                  'null' + '-' + 'level' + constants.CSLE.NAME + '--1') is True
        assert ContainerController.stop_container("John Doe") is False

    def test_rm_all_stopped_containers(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture):
        """
        Tests the rm_all_stopped_containers in the ContainerController

        :param mocker: the Pytest mocker object
        :param client_1: fixture for mocking the docker.from_env class
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        assert ContainerController.rm_all_stopped_containers() is None

    def test_rm_container(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture) -> None:
        """
        Tests the rm_container method in the ContainerController
        
        :param mocker: the Pytest mocker object
        :param client_1: fixture for mocking the docker.from_env class
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        assert ContainerController.rm_container(constants.CONTAINER_IMAGES.CSLE_PREFIX +
                                                'null' + '-' + 'level' + constants.CSLE.NAME + '--1') is True
        assert ContainerController.rm_container("JohnDoe") is False

    def test_rm_all_images(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture) -> None:
        """
        Tests the rm_all_images method in the ContainerController
        
        :param mocker: the Pytest mocker object
        :param client_1: fixture for mocking the docker.from_env class
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        assert ContainerController.rm_all_images() is None

    def test_rm_image(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture) -> None:
        """
        Tests the rm_image method in the ContainerController
        
        :param mocker: the Pytest mocker object
        :param client_1: fixture for mocking the docker.from_env class
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        assert ContainerController.rm_image(constants.CSLE.NAME) is True
        assert ContainerController.rm_image("JDoeName") is False

    def test_list_all_images(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture) -> None:
        """
        Tests the list_all_images method inte the ContainerController

        :param mocker: the Pytest mocker object
        :param client_1: fixture for mocking the ContainerController
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        images_names = ContainerController.list_all_images()[0]
        assert images_names[0] == constants.CSLE.NAME
        assert images_names[1] == "yes"
        assert images_names[2] == "os"
        assert images_names[3] == "csle-architecture"
        assert images_names[4] == 100

    def test_list_docker_networks(self, mocker: pytest_mock.MockFixture, file_opener: pytest_mock.MockFixture) -> None:
        """
        Tests the list_docker_networks method inte the ContainerController

        :param mocker: the Pytest mocker object
        :param file_opener: fixture for mocking the ContainerController
        :return: None
        """
        mocker.patch('os.popen', side_effect=file_opener)
        test_networks_ids_str, test_network_ids = ContainerController.list_docker_networks()
        assert test_networks_ids_str[0] == f"{constants.CSLE.CSLE_NETWORK_PREFIX}1"
        assert test_networks_ids_str[1] == f"{constants.CSLE.CSLE_NETWORK_PREFIX}2"
        assert test_network_ids[0] == 1
        assert test_network_ids[1] == 2

    def test_list_all_networks(self, mocker: pytest_mock.MockFixture, file_opener: pytest_mock.MockFixture) -> None:
        """
        Tests the list_all_networks method
        
        :param mocker: the pytest mocker object
        :param file_opener: fixture for mocking the ContainerController
        :return: None
        """
        mocker.patch('os.popen', side_effect=file_opener)
        test_networks_ids_str = ContainerController.list_all_networks()
        assert test_networks_ids_str[0] == f"{constants.CSLE.CSLE_NETWORK_PREFIX}1"
        assert test_networks_ids_str[1] == f"{constants.CSLE.CSLE_NETWORK_PREFIX}2"

    def test_start_all_stopped_containers(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture) \
            -> None:
        """
        Testing the start_all_stopped_containers method in the ContainerController

        :param mocker: the pytest mocker object
        :param client_1: pytest fixture for mocking the ContainerController
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        assert ContainerController.start_all_stopped_containers() is None

    def test_start_container(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture) -> None:
        """
        Testing the start_container method in the ContainerController
        
        :param mocker: the Pytest mocker object
        :param client_1: pytest fixture for mocking the ContainerController
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        assert ContainerController.start_container(constants.CONTAINER_IMAGES.CSLE_PREFIX +
                                                   'null' + '-' + 'level' + constants.CSLE.NAME + '--1') is True
        assert ContainerController.start_container("JohnDoe") is False

    def test_list_all_running_containers(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture,
                                         client_2: pytest_mock.MockFixture,
                                         example_docker_env_metadata: DockerEnvMetadata) -> None:
        """
        Testing the list_all_running_containers method in the ContainerController

        :param mocker: the pytest mocker object
        :param client_1: pytest fixture for mocking the docker client 1
        :param client_2: pytest fixture for mocking the docker client 2
        :param example_docker_env_metadata: example docker env metadata
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.APIClient', side_effect=client_2)
        mocker.patch('csle_common.util.docker_util.DockerUtil.parse_runnning_emulation_infos',
                     result=[example_docker_env_metadata])
        assert len(ContainerController.list_all_running_containers()) == 0

    def test_list_all_running_containers_in_emulation(self, mocker: pytest_mock.MockFixture,
                                                      example_emulation_env_config, client_1: pytest_mock.MockFixture,
                                                      client_2: pytest_mock.MockFixture,
                                                      example_docker_env_metadata: DockerEnvMetadata) -> None:
        """
        Testing the list_all_running_containers_in_emulation in the ContainerController

        :param mocker: the pytest mocker object
        :param example_emulation_env_config: an example object being the emulation environmnt configuration
        :param client_1: pytest fixture for mocking the Docker first client
        :param client_2: pytest fixture for mocking the Docker second client
        :param example_docker_env_metadata: example docker env metadata
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.APIClient', side_effect=client_2)
        mocker.patch('csle_common.util.docker_util.DockerUtil.parse_runnning_emulation_infos',
                     result=[example_docker_env_metadata])
        running_emulation_containers, stopped_emulation_containers = ContainerController. \
            list_all_running_containers_in_emulation(example_emulation_env_config)
        assert len(running_emulation_containers) == 0
        assert stopped_emulation_containers[0].to_dict() == example_emulation_env_config. \
            containers_config.containers[0].to_dict()
        assert stopped_emulation_containers[1].to_dict() == example_emulation_env_config. \
            kafka_config.container.to_dict()
        assert stopped_emulation_containers[2].to_dict() == \
               example_emulation_env_config.elk_config.container.to_dict()
        assert stopped_emulation_containers[3].to_dict() == \
               example_emulation_env_config.sdn_controller_config.container.to_dict()

    def test_list_all_active_networks_for_emulation(self, mocker: pytest_mock.MockFixture,
                                                    example_emulation_env_config: EmulationEnvConfig,
                                                    file_opener) -> None:
        """
        Testing the list_all_active_networks_for_emulation in the ContainerController

        :param mocker: the pytest mocker object
        :param example_emulation_env_config: an example object being the emulation environment configuration
        :param file_opener: fixture for mocking the ContainerController
        :return: None
        """
        mocker.patch("os.popen", side_effect=file_opener)
        active_emulation_networks, inactive_emulation_networks = \
            ContainerController.list_all_active_networks_for_emulation(
                example_emulation_env_config)
        assert inactive_emulation_networks == []
        assert active_emulation_networks[0].to_dict() == \
               example_emulation_env_config.containers_config.networks[0].to_dict()

    def test_list_running_emulations(self, mocker: pytest_mock.MockFixture,
                                     example_emulation_env_config: EmulationEnvConfig,
                                     client_1: pytest_mock.MockFixture,
                                     client_2: pytest_mock.MockFixture,
                                     example_docker_env_metadata: DockerEnvMetadata) -> None:
        """
        Testing the list_running_emulations method in the ContainerController
        
        :param mocker:the pytest mocker object
        :param example_emulation_env_config: fixture for mocking, fetched from the conftest file
        :param client_1: pytest fixture for mocking the Docker first client
        :param client_2: pytest fixture for mocking the Docker second client
        :param example_docker_env_metadata: example docker env metadata
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.APIClient', side_effect=client_2)
        mocker.patch('csle_common.util.docker_util.DockerUtil.parse_runnning_emulation_infos',
                     result=[example_docker_env_metadata])
        emulation_name_list = ContainerController.list_running_emulations()
        assert len(emulation_name_list) == 0

    def test_is_emulation_running(self, mocker: pytest_mock.MockFixture,
                                  example_emulation_env_config: EmulationEnvConfig,
                                  client_1: pytest_mock.MockFixture,
                                  client_2: pytest_mock.MockFixture,
                                  example_docker_env_metadata: DockerEnvMetadata) -> None:
        """
        Testing the is_emulation_running method in the ContainerController

        :param mocker:the pytest mocker object
        :param example_emulation_env_config: fixture for mocking, fetched from the conftest file
        :param client_1: pytest fixture for mocking the Docker first client
        :param client_2: pytest fixture for mocking the Docker second client
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.APIClient', side_effect=client_2)
        mocker.patch('csle_common.util.docker_util.DockerUtil.parse_runnning_emulation_infos',
                     result=[example_docker_env_metadata])
        assert not ContainerController.is_emulation_running(example_emulation_env_config)

    def test_list_all_stopped_containers(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture,
                                         client_2: pytest_mock.MockFixture) -> None:
        """
        Testing the is_emulation_running method in the ContainerController

        :param mocker:the pytest mocker object
        :param example_emulation_env_config: fixture for mocking, fetched from the conftest file
        :param client_1: pytest fixture for mocking the Docker first client
        :param client_2: pytest fixture for mocking the Docker second client
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.APIClient', side_effect=client_2)
        test_tuple = ContainerController.list_all_stopped_containers()[0]
        assert test_tuple[0] == constants.CONTAINER_IMAGES.CSLE_PREFIX + \
               'null' + '-' + 'level' + constants.CSLE.NAME + '--1'
        assert test_tuple[1] == 'JDoeImage'
        assert test_tuple[2] == '123.456.78.99'

    def test_get_network_references(self, mocker: pytest_mock.MockFixture, client_1: pytest_mock.MockFixture) -> None:
        """
        Testing the get_network_references method in the ContainerController

        :param mocker:the pytest mocker object
        :param client_1: pytest fixture for mocking the Docker first client
        :param client_2: pytest fixture for mocking the Docker second client
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        test_networks = ContainerController.get_network_references()
        assert test_networks[0].name == constants.CONTAINER_IMAGES.CSLE_PREFIX + \
               'null' + '-' + 'level' + constants.CSLE.NAME + '--1'

    def test_create_networks(self, mocker: pytest_mock.MockFixture, client_1,
                             example_containers_config: ContainersConfig, example_config: Config, ipam_pool,
                             ipam_config) -> None:
        """
        Testing the create_networks method in the ContainerController

        :param mocker: the pytest mocker object
        :param client_1: pytest fixture for mocking the Docker first client
        :param client_2: pytest fixture for mocking the Docker second client
        :param example_containers_config: example containers config
        :param example_config: example config DTO
        :param ipam_pool: fixture for the Docker ipam_pool
        :param ipam_config: fixture for the Docker ipam_config
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('csle_common.dao.emulation_config.config.Config.get_current_config', result=example_config)
        mocker.patch('docker.types.IPAMPool', side_effect=ipam_pool)
        mocker.patch('docker.types.IPAMConfig', side_effect=ipam_config)
        creation = ContainerController.create_networks(example_containers_config, logging.Logger("test"))
        assert creation is None

    def test_connect_containers_to_networks(self, mocker: pytest_mock.MockFixture,
                                            example_emulation_env_config: EmulationEnvConfig,
                                            sub_popen, client_1, client_2, file_opener) -> None:
        """
        Testing the connect_containers_to_networks method

        :param mocker: the pytest mocker object
        :param example_emulation_env_config: EmulationEnvironmentConfig fetched from the conftest file
        :param sub_popen: fixture mocking the subprocess.Popen
        :param client_1: pytest fixture for mocking the Docker first client
        :param client_2: pytest fixture for mocking the Docker second client
        :return: None
        """
        mocker.patch('time.sleep', return_value=None)
        mocker.patch('subprocess.Popen', side_effect=sub_popen)
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.APIClient', side_effect=client_2)
        mocker.patch('os.popen', side_effect=file_opener)
        connection = ContainerController.connect_containers_to_networks(example_emulation_env_config, '123.456.78.99',
                                                                        logger=logging.Logger("test"))
        assert connection is None

    def test_connect_container_to_network(self, mocker: pytest_mock.MockFixture,
                                          example_node_container_config: NodeContainerConfig, sub_popen,
                                          file_opener, client_1, client_2, ) -> None:
        """
        Testing connect_container_to_network method in the ContainerController
        
        :param mocker: the pytest mocker object
        :param example_node_container_config: an example NodeContainerConfig fetched from the conftest file
        :param sub_popen: fixture mocking the subprocess.Popen
        :param file_opener: fixture mocking the os.popen
        :return: None
        """
        mocker.patch('time.sleep', return_value=None)
        mocker.patch('subprocess.Popen', side_effect=sub_popen)
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.APIClient', side_effect=client_2)
        mocker.patch('os.popen', side_effect=file_opener)
        connection = ContainerController.connect_container_to_network(
            example_node_container_config, logging.Logger("test"))
        assert connection is None

    def test_start_docker_stats_thread(self, mocker: pytest_mock.MockFixture,
                                       example_emulation_execution: EmulationExecution, true_running_stats_manager,
                                       false_running_stats_manager, sub_popen, start_fixture) -> None:
        """
        Testing the start_docker_stats_thread in the ContainerController

        :param mocker: the pytest mocker object
        :param example_node_container_config: an example NodeContainerConfig fetched from the conftest file
        :param sub_popen: fixture mocking the subprocess.Popen
        :param file_opener: fixture mocking the os.popen
        :return: None
        """
        mocker.patch('time.sleep', return_value=None)
        mocker.patch('subprocess.Popen', side_effect=sub_popen)
        mocker.patch('csle_common.controllers.management_system_controller.' +
                     'ManagementSystemController.is_statsmanager_running',
                     side_effect=true_running_stats_manager)
        mocker.patch('csle_collector.docker_stats_manager.query_docker_stats_manager.start_docker_stats_monitor',
                     side_effect=start_fixture)
        starter = ContainerController.start_docker_stats_thread(example_emulation_execution,
                                                                physical_server_ip='123.456.78.99',
                                                                logger=logging.Logger("test"))
        assert starter is None
        mocker.patch('csle_common.controllers.management_system_controller.' +
                     'ManagementSystemController.is_statsmanager_running',
                     side_effect=false_running_stats_manager)
        ContainerController.start_docker_stats_thread(example_emulation_execution,
                                                      physical_server_ip='123.456.78.99',
                                                      logger=logging.Logger("test"))

    def test_stop_docker_stats_thread(
            self, mocker: pytest_mock.MockFixture, example_emulation_execution: EmulationExecution, true_stop_fixture,
            false_running_stats_manager, true_running_stats_manager, false_stop_fixture, stop_fixture) -> None:
        """
        Testing the stop_docker_stats_thread

        :param mocker: the pytest mocker object
        :param example_emulation_execution: EmulationExecution object fetched from the config file
        :param true_stop_fixture: pytest fixture
        :param false_running_stats_manager: pytest fixture
        :param true_running_stats_manager: pytest fixture
        :param false_stop_fixture: pytest fixture
        :param stop_fixture: pytest fixture
        :return: None
        """
        mocker.patch('time.sleep', return_value=None)
        mocker.patch('csle_collector.docker_stats_manager.query_docker_stats_manager.stop_docker_stats_monitor',
                     side_effect=stop_fixture)
        mocker.patch('csle_common.controllers.management_system_controller.' +
                     'ManagementSystemController.is_statsmanager_running',
                     side_effect=true_running_stats_manager)
        mocker.patch('csle_common.controllers.management_system_controller.' +
                     'ManagementSystemController.stop_docker_statsmanager',
                     side_effect=true_stop_fixture)
        stopper = ContainerController.stop_docker_stats_thread(example_emulation_execution, "123.456.78.99",
                                                               logging.getLogger("test"))
        assert stopper is None
        mocker.patch('csle_common.controllers.management_system_controller.' +
                     'ManagementSystemController.stop_docker_statsmanager',
                     side_effect=false_stop_fixture)
        stopper = ContainerController.stop_docker_stats_thread(example_emulation_execution, "123.456.78.99",
                                                               logging.getLogger("test"))
        assert stopper is None
        mocker.patch('csle_common.controllers.management_system_controller.' +
                     'ManagementSystemController.is_statsmanager_running',
                     side_effect=false_running_stats_manager)
        stopper = ContainerController.stop_docker_stats_thread(example_emulation_execution, "123.456.78.99",
                                                               logging.getLogger("test"))
        assert stopper is None

    def test_get_docker_stats_manager_status(
            self, mocker: pytest_mock.MockFixture, host_ip,
            example_docker_stats_manager_config: DockerStatsManagerConfig, dsm_status, stub) -> None:
        """
        Testing the get_docker_stats_manager_status method in the ContainerController

        :param mocker: the pytest mocker object
        :param host_ip: the host_ip fixture
        :param example_docker_stats_manager_config: the example_docker_stats_manager_config fixture
        :param dsm_status: the dsm_status fixture
        :param stub: the stub fixture
        :return: None
        """
        mocker.patch('csle_common.util.general_util.GeneralUtil.get_host_ip', side_effect=host_ip)
        mocker.patch('csle_collector.docker_stats_manager.query_docker_stats_manager.get_docker_stats_manager_status',
                     side_effect=dsm_status)
        mocker.patch('csle_collector.docker_stats_manager.' +
                     'docker_stats_manager_pb2_grpc.DockerStatsManagerStub', side_effect=stub)
        dsm_dto = ContainerController.get_docker_stats_manager_status(example_docker_stats_manager_config)
        assert dsm_dto.num_monitors == 10
        assert dsm_dto.emulations == ["JDoeEmulation"]
        assert dsm_dto.emulation_executions == [10]

    def test_get_docker_stats_manager_status_by_ip_and_port(
            self, mocker: pytest_mock.MockFixture, host_ip, dsm_status, stub) -> None:
        """
        Testing the get_docker_stats_manager_status_by_ip_and_port method in the ContainerController

        :param mocker: the pytest mocker object
        :param host_ip: the host_ip fixture
        :param dsm_status: the dsm_status fixture
        :param stub: the stub fixture
        :return: None
        """
        mocker.patch('csle_common.util.general_util.GeneralUtil.get_host_ip', side_effet=host_ip)
        mocker.patch('csle_collector.docker_stats_manager.query_docker_stats_manager.get_docker_stats_manager_status',
                     side_effect=dsm_status)
        mocker.patch('csle_collector.docker_stats_manager.docker_stats_manager_pb2_grpc' +
                     '.DockerStatsManagerStub', side_effect=stub)
        dsm_dto = ContainerController.get_docker_stats_manager_status_by_ip_and_port("123.456.78.99", 1)
        assert dsm_dto.num_monitors == 10
        assert dsm_dto.emulations == ["JDoeEmulation"]
        assert dsm_dto.emulation_executions == [10]

    def test_create_network_from_dto(self, mocker: pytest_mock.MockFixture, example_container_network: ContainerNetwork,
                                     client_1, ipam_pool, ipam_config, example_config: Config) -> None:
        """
        Testing the create_network_from_dto in the ContainerController

        :param mocker: the pytest mocker object
        :param client_1: the client_1 fixture
        :param ipam_pool: the ipam_pool fixture
        :param ipam_config: the ipam_confoig fixture
        :param example_config: the example_config fixture from conftest
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.types.IPAMPool', side_effect=ipam_pool)
        mocker.patch('docker.types.IPAMConfig', side_effect=ipam_config)
        mocker.patch('csle_common.dao.emulation_config.config.Config.get_current_config', result=example_config)
        creator = ContainerController.create_network_from_dto(example_container_network, logging.Logger("test"))
        assert creator is None

    def test_create_network(self, mocker: pytest_mock.MockFixture, client_1, ipam_pool, ipam_config) -> None:
        """
        Testing the create_network in the ContainerController

        :param mocker: the pytest mocker object
        :param client_1: the client_1 fixture
        :param ipam_pool: the ipam_pool fixture
        :param ipam_config: the ipam_confoig fixture
        :param example_config: the example_config fixture from conftest
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.types.IPAMPool', side_effect=ipam_pool)
        mocker.patch('docker.types.IPAMConfig', side_effect=ipam_config)
        creator = ContainerController.create_network(name="JDoeCreator", subnetmask="null",
                                                     logger=logging.Logger("test"))
        assert creator is None

    def test_remove_network(self, mocker: pytest_mock.MockFixture, client_1) -> None:
        """
        Testing the remove_network method in the ContainerController

        :param mocker: the pytest mocker object
        :param client_1: the client_1 fixture
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        remover = ContainerController.remove_network(constants.CONTAINER_IMAGES.CSLE_PREFIX +
                                                     'null' + '-' + 'level' + constants.CSLE.NAME + '--1',
                                                     logging.getLogger())
        assert remover is None

    def test_remove_networks(self, mocker: pytest_mock.MockFixture, client_1) -> None:
        """
        Testing the remove_network method in the ContainerController

        :param mocker: the pytests mocker object
        :param client_1: the client_1 fixture
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        remover = ContainerController.remove_networks([constants.CONTAINER_IMAGES.CSLE_PREFIX +
                                                       'null' + '-' + 'level' + constants.CSLE.NAME + '--1'],
                                                      logging.getLogger())
        assert remover is True
        remover = ContainerController.remove_networks(["JohnDoe"], logging.getLogger())
        assert remover is False

    def test_rm_all_networks(self, mocker: pytest_mock.MockFixture, client_1) -> None:
        """
        Testing the rm_all_networks method in the ContainerController
        :param mocker_ the pytest mocker object
        :param client_1: the client_1 fixture
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        remover = ContainerController.rm_all_networks(logging.getLogger())
        assert remover is None

    def test_rm_network(self, mocker: pytest_mock.MockFixture, client_1) -> None:
        """
        Testing the rm_network method in the ContainerController

        :param mocker: the pytest mocker object
        :param client_1: the client_1 fixture
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        remover = ContainerController.rm_network(constants.CSLE.NAME, logging.getLogger())
        assert remover is True
        remover = ContainerController.rm_network("JDoe", logging.getLogger())
        assert remover is False

    def test_run_command(self, mocker: pytest_mock.MockFixture, client_1, client_2) -> None:
        """
        Testing the run_command method in the ContainerController
        
        :param mocker: the pytest mocker object
        :param client_1: the client_1 fixture
        :param client_2: the client_2 fixture
        :return: None
        """
        mocker.patch('docker.from_env', side_effect=client_1)
        mocker.patch('docker.APIClient', side_effect=client_2)
        runner = ContainerController.run_command(cmd=constants.MANAGEMENT.LIST_STOPPED)
        assert runner is None

    def test_get_docker_stats_managers_ips(
            self, mocker: pytest_mock.MockFixture, socket_fix, example_emulation_env_config: EmulationEnvConfig) \
            -> None:
        """
        Testing the get_docker_stats_managers_ips
        
        :param mocker: the pytest mocker object
        :param host_ip: the host_ip fixture
        :param example_emulation_env_config: the example_emulation_env_config fixture from the conftest
        :return: None
        """
        mocker.patch('socket.socket', side_effect=socket_fix)
        ip = ContainerController.get_docker_stats_managers_ips(example_emulation_env_config)
        assert ip == ["123.456.78.99"]

    def test_get_docker_stats_managers_ports(self, example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Testing the get_docker_stats_managers_ports in the ContainerController

        :param mocker: the pytest mocker object
        :pram example_emulation_config: the example_emulation_config fixture fetched from the conftest file
        :return: None
        """
        test_list = ContainerController.get_docker_stats_managers_ports(example_emulation_env_config)
        assert test_list[0] == 50046

    def test_get_docker_stats_managers_info(
            self, mocker: pytest_mock.MockFixture, socket_fix, example_emulation_env_config: EmulationEnvConfig,
            host_ip, dsm_status, stub, example_docker_stats_managers_info: DockerStatsManagersInfo) -> None:
        """
        Testing the get_docker_stats_managers_info
        
        :param mocker: thepytest mocker object
        :param socket_fix: the socket_fix fixture
        :pram example_emulation_config: the example_emulation_config fixture fetched from the conftest file
        :param host_ip: the host_ip fixture
        :param dsm_status: the dsm_status fixture
        :param stub: the stub fixture
        :return: None
        """
        mocker.patch('time.sleep', return_value=None)
        mocker.patch('socket.socket', side_effect=socket_fix)
        mocker.patch('csle_collector.docker_stats_manager.query_docker_stats_manager.get_docker_stats_manager_status',
                     side_effect=dsm_status)
        mocker.patch('csle_collector.docker_stats_manager.docker_stats_manager_pb2_grpc.DockerStatsManagerStub',
                     side_effect=stub)
        docker_stat = ContainerController.get_docker_stats_managers_info(
            example_emulation_env_config, ["123.456.78.99"], "123.456.78.99", logging.getLogger())
        assert docker_stat.docker_stats_managers_running[0] is True
        assert docker_stat.ips[0] == "123.456.78.99"
        assert docker_stat.emulation_name == "JDoeEmulation"
        assert docker_stat.execution_id == 10
        docker_stat_dict = docker_stat.to_dict()
        assert docker_stat_dict["docker_stats_managers_statuses"][0]["num_monitors"] == 10
        assert docker_stat_dict["docker_stats_managers_statuses"][0]["emulations"][0] == "JDoeEmulation"
        assert docker_stat_dict["docker_stats_managers_statuses"][0]["emulation_executions"][0] == 10
        assert docker_stat.ports[0] == 50046



from typing import List, Dict, Any, Union
from csle_common.dao.docker.docker_container_metadata import DockerContainerMetadata
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.kafka_config import KafkaConfig
from csle_base.json_serializable import JSONSerializable


class DockerEnvMetadata(JSONSerializable):
    """
    DTO Object representing a running environment
    """

    def __init__(self, containers: List[DockerContainerMetadata], name: str, subnet_prefix: str,
                 subnet_mask: str, level: str, config: Union[None, EmulationEnvConfig],
                 kafka_config: Union[None, KafkaConfig]):
        """
        Initializes the DTO

        :param containers: the list of running containers
        :param name: the environment name
        :param subnet_prefix: the subnet prefix
        :param subnet_mask: the subnet mask
        :param level: the level of the environment
        :param config: the configuration of the environment
        :param kafka_config: the kafka configuration
        """
        self.containers = containers
        self.name = name
        self.subnet_prefix = subnet_prefix
        self.subnet_mask = subnet_mask
        self.level = level
        self.config = config
        self.kafka_config = kafka_config

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DockerEnvMetadata":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        config = None
        kafka_config = None
        if d["config"] is not None:
            config = EmulationEnvConfig.from_dict(d["config"])
        if d["kafka_config"] is not None:
            config = KafkaConfig.from_dict(d["kafka_config"])
        obj = DockerEnvMetadata(
            containers=list(map(lambda x: DockerContainerMetadata.from_dict(x), d["containers"])),
            name=d["name"], subnet_prefix=d["subnet_prefix"], subnet_mask=d["subnet_mask"],
            level=d["level"], config=config, kafka_config=kafka_config)
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["containers"] = list(map(lambda x: x.to_dict(), self.containers))
        d["name"] = self.name
        d["subnet_prefix"] = self.subnet_prefix
        d["subnet_mask"] = self.subnet_mask
        d["num_containers"] = len(self.containers)
        d["level"] = self.level
        if self.config is not None:
            d["config"] = self.config.to_dict()
        else:
            d["config"] = None
        if self.kafka_config is not None:
            d["kafka_config"] = self.kafka_config.to_dict()
        else:
            d["kafka_config"] = None
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "DockerEnvMetadata":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return DockerEnvMetadata.from_dict(json.loads(json_str))



from typing import Dict, Any
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.credential import Credential
from csle_base.json_serializable import JSONSerializable


class NmapBruteCredentials(JSONSerializable):
    """
    A DTO representing credentials found with NMAP Brute-Force Scans
    """

    def __init__(self, username: str, pw: str, state: str, port: int, protocol: TransportProtocol, service: str):
        """
        Initializes the DTO

        :param username: the username of the credential
        :param pw: the pw of the credential
        :param state: the state of the credential
        :param port: the port of the scan
        :param protocol: the protocol of the scan
        :param service: the service of the credential
        """
        self.username = username
        self.pw = pw
        self.state = state
        self.port = port
        self.protocol = protocol
        self.service = service

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"username:{self.username},pw:{self.pw},state:{self.statef},port:{self.port},protocol:{self.protocol}," \
               f"service:{self.service}"

    def to_obs(self) -> Credential:
        """
        Converts the Object into a Credential Object

        :return: the created Credential Object
        """
        return Credential(username=self.username, pw=self.pw, port=int(self.port), service=self.service,
                          protocol=self.protocol)

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["username"] = self.username
        d["pw"] = self.pw
        d["state"] = self.state
        d["port"] = self.port
        d["protocol"] = self.protocol
        d["service"] = self.service
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapBruteCredentials":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapBruteCredentials(username=d["username"], pw=d["pw"], state=d["state"], port=d["port"],
                                   protocol=d["protocol"], service=d["service"])
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapBruteCredentials":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapBruteCredentials.from_dict(json.loads(json_str))

    def __hash__(self) -> int:
        """
        :return: a hash representation of the object
        """
        return (hash(self.username) + 31 * hash(self.pw) + 31 * hash(self.state) + 28 * hash(self.service)
                + 38 * (self.protocol.value))



import numpy as np
import random
import csle_collector.constants.constants as constants
from csle_collector.host_manager.dao.host_metrics import HostMetrics
from csle_common.consumer_threads.host_metrics_consumer_thread import HostMetricsConsumerThread
from csle_common.consumer_threads.aggregated_host_metrics_thread import AggregatedHostMetricsThread
from csle_collector.ossec_ids_manager.dao.ossec_ids_alert_counters import OSSECIdsAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_alert_counters import SnortIdsAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_rule_counters import SnortIdsRuleCounters
from csle_common.consumer_threads.aggregated_ossec_ids_log_consumer_thread import AggregatedOSSECIdsLogConsumerThread
from csle_common.consumer_threads.aggregated_snort_ids_log_consumer_thread import AggregatedSnortIdsLogConsumerThread
from csle_common.consumer_threads.aggregated_snort_ids_rule_log_consumer_thread \
    import AggregatedSnortIdsRuleLogConsumerThread
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.consumer_threads.attacker_actions_consumer_thread import AttackerActionsConsumerThread
from csle_common.dao.emulation_observation.defender.emulation_defender_machine_observation_state import \
    EmulationDefenderMachineObservationState
from csle_collector.client_manager.client_population_metrics import ClientPopulationMetrics
from csle_collector.docker_stats_manager.dao.docker_stats import DockerStats
from csle_collector.snort_ids_manager.dao.snort_ids_ip_alert_counters import SnortIdsIPAlertCounters
from csle_common.consumer_threads.avg_host_metrics_thread import AvgHostMetricsThread
from csle_common.consumer_threads.client_population_consumer_thread import ClientPopulationConsumerThread
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_action.defender.emulation_defender_action_id import EmulationDefenderActionId
from csle_common.dao.emulation_action.defender.emulation_defender_action_type import EmulationDefenderActionType
from csle_common.consumer_threads.defender_actions_consumer_thread import DefenderActionsConsumerThread
from csle_common.consumer_threads.docker_host_stats_consumer_thread import DockerHostStatsConsumerThread
from csle_common.consumer_threads.docker_stats_consumer_thread import DockerStatsConsumerThread
from csle_common.consumer_threads.ossec_ids_log_consumer_thread import OSSECIdsLogConsumerThread
from csle_common.consumer_threads.snort_ids_log_consumer_thread import SnortIdsLogConsumerThread


class TestConsumerThreadsSuiteSuite:
    """
    Test suite for consumer threads
    """

    def test_host_metrics_consumer_thread(self) -> None:
        """
        Tests creation of a host metrics consumer thread and its methods

        :return: None
        """
        example_host_metrics = HostMetrics(num_logged_in_users=1, num_failed_login_attempts=10,
                                           num_open_connections=1, num_login_events=2,
                                           num_processes=5, num_users=1, ip="1.1.1.1", ts=0.5)
        thread = HostMetricsConsumerThread(host_ip="8.8.8.8", kafka_server_ip="8.8.8.8", kafka_port=3030,
                                           host_metrics=example_host_metrics)
        assert thread.host_metrics == example_host_metrics

    def test_aggregated_host_metrics_consumer_thread(self) -> None:
        """
        Tests creation of the aggregated host metrics consumer thread and its methods

        :return: None
        """
        example_host_metrics = HostMetrics(num_logged_in_users=1, num_failed_login_attempts=10,
                                           num_open_connections=1, num_login_events=2,
                                           num_processes=5, num_users=1, ip="1.1.1.1", ts=0.5)
        thread = AggregatedHostMetricsThread(host_metrics=example_host_metrics, machines=[], sleep_time=10)
        assert thread.host_metrics == example_host_metrics
        assert thread.sleep_time == 10
        assert len(thread.machines) == 0

        # Test empty list
        thread.host_metrics_list = []
        averaged_host_metrics = thread.get_average_aggregated_host_metrics()
        assert averaged_host_metrics == example_host_metrics

        # Test singleton list
        thread.host_metrics_list = [example_host_metrics]
        averaged_host_metrics = thread.get_average_aggregated_host_metrics()
        assert averaged_host_metrics == example_host_metrics

        # Test list with 10 identical objects
        thread.host_metrics_list = [example_host_metrics] * 10
        averaged_host_metrics = thread.get_average_aggregated_host_metrics()
        assert averaged_host_metrics.num_logged_in_users == example_host_metrics.num_logged_in_users
        assert averaged_host_metrics.num_failed_login_attempts == example_host_metrics.num_failed_login_attempts
        assert averaged_host_metrics.num_open_connections == example_host_metrics.num_open_connections
        assert averaged_host_metrics.num_login_events == example_host_metrics.num_login_events
        assert averaged_host_metrics.num_processes == example_host_metrics.num_processes
        assert averaged_host_metrics.num_users == example_host_metrics.num_users

        # Test list with two different objects
        example_host_metrics_2 = HostMetrics(num_logged_in_users=2, num_failed_login_attempts=5,
                                             num_open_connections=7, num_login_events=9,
                                             num_processes=11, num_users=3, ip="1.1.1.1", ts=0.5)
        thread.host_metrics_list = [example_host_metrics, example_host_metrics_2]
        averaged_host_metrics = thread.get_average_aggregated_host_metrics()
        assert (averaged_host_metrics.num_logged_in_users ==
                int(round((example_host_metrics.num_logged_in_users + example_host_metrics_2.num_logged_in_users) / 2)))
        assert (averaged_host_metrics.num_failed_login_attempts ==
                int(round((example_host_metrics.num_failed_login_attempts +
                           example_host_metrics_2.num_failed_login_attempts) / 2)))
        assert (averaged_host_metrics.num_open_connections ==
                int(round(
                    (example_host_metrics.num_open_connections + example_host_metrics_2.num_open_connections) / 2)))
        assert (averaged_host_metrics.num_login_events ==
                int(round((example_host_metrics.num_login_events + example_host_metrics_2.num_login_events) / 2)))
        assert (averaged_host_metrics.num_processes ==
                int(round((example_host_metrics.num_processes + example_host_metrics_2.num_processes) / 2)))
        assert (averaged_host_metrics.num_users ==
                int(round((example_host_metrics.num_users + example_host_metrics_2.num_users) / 2)))

    def test_aggregated_ossec_ids_log_consumer_thread(self) -> None:
        """
        Tests creation of a ossec ids log consumer thread and its methods

        :param example_ossec_ids_alert_counters: an object of OSSECIdsAlertCounters
        :return: None
        """
        example_ossec_ids_alert_counters = OSSECIdsAlertCounters()
        example_ossec_ids_alert_counters.level_alerts = list(np.zeros(16))
        for idx in range(16):
            example_ossec_ids_alert_counters.level_alerts[idx] = random.randint(0, 10)

        example_ossec_ids_alert_counters.group_alerts = list(
            np.zeros(len(set(constants.OSSEC.OSSEC_IDS_ALERT_GROUP_ID.values()))))
        for idx in range(len(example_ossec_ids_alert_counters.group_alerts)):
            example_ossec_ids_alert_counters.group_alerts[idx] = random.randint(0, 10)

        example_ossec_ids_alert_counters.severe_alerts = 10
        example_ossec_ids_alert_counters.warning_alerts = 4
        example_ossec_ids_alert_counters.total_alerts = 21
        example_ossec_ids_alert_counters.alerts_weighted_by_level = 0.3

        thread = AggregatedOSSECIdsLogConsumerThread(
            kafka_server_ip="1.2.3.4", kafka_port=1234, ossec_ids_alert_counters=example_ossec_ids_alert_counters.copy()
        )
        assert thread.ossec_ids_alert_counters == example_ossec_ids_alert_counters
        assert thread.kafka_port == 1234
        assert thread.kafka_server_ip == "1.2.3.4"

        # Test empty list
        thread.ossec_ids_alert_counters_list = []
        aggregated_ossec_ids_alert_counters = thread.get_aggregated_ids_alert_counters()
        assert aggregated_ossec_ids_alert_counters == example_ossec_ids_alert_counters

        # Test singleton list
        thread.ossec_ids_alert_counters_list = [example_ossec_ids_alert_counters.copy()]
        aggregated_ossec_ids_alert_counters = thread.get_aggregated_ids_alert_counters()
        assert aggregated_ossec_ids_alert_counters == example_ossec_ids_alert_counters

        # Test list with 10 identical objects
        thread.ossec_ids_alert_counters_list = []
        for i in range(10):
            thread.ossec_ids_alert_counters_list.append(example_ossec_ids_alert_counters.copy())
        aggregated_ossec_ids_alert_counters = thread.get_aggregated_ids_alert_counters()
        assert (aggregated_ossec_ids_alert_counters.severe_alerts ==
                10 * example_ossec_ids_alert_counters.severe_alerts)
        assert (aggregated_ossec_ids_alert_counters.warning_alerts ==
                10 * example_ossec_ids_alert_counters.warning_alerts)
        assert (aggregated_ossec_ids_alert_counters.total_alerts == 10 * example_ossec_ids_alert_counters.total_alerts)
        assert (round(aggregated_ossec_ids_alert_counters.alerts_weighted_by_level, 3) ==
                round(10 * example_ossec_ids_alert_counters.alerts_weighted_by_level, 3))
        for idx in range(len(aggregated_ossec_ids_alert_counters.level_alerts)):
            assert (aggregated_ossec_ids_alert_counters.level_alerts[idx] ==
                    10 * example_ossec_ids_alert_counters.level_alerts[idx])
        for idx in range(len(aggregated_ossec_ids_alert_counters.group_alerts)):
            assert (aggregated_ossec_ids_alert_counters.group_alerts[idx] ==
                    10 * example_ossec_ids_alert_counters.group_alerts[idx])

        # Test list with two different objects
        example_ossec_ids_alert_counters_2 = OSSECIdsAlertCounters()
        example_ossec_ids_alert_counters_2.level_alerts = list(np.zeros(16))
        for idx in range(16):
            example_ossec_ids_alert_counters_2.level_alerts[idx] = random.randint(0, 10)

        example_ossec_ids_alert_counters_2.group_alerts = list(
            np.zeros(len(set(constants.OSSEC.OSSEC_IDS_ALERT_GROUP_ID.values()))))
        for idx in range(len(example_ossec_ids_alert_counters_2.group_alerts)):
            example_ossec_ids_alert_counters_2.group_alerts[idx] = random.randint(0, 10)

        example_ossec_ids_alert_counters_2.severe_alerts = 15
        example_ossec_ids_alert_counters_2.warning_alerts = 1
        example_ossec_ids_alert_counters_2.total_alerts = 18
        example_ossec_ids_alert_counters_2.alerts_weighted_by_level = 0.7

        thread.ossec_ids_alert_counters_list = [example_ossec_ids_alert_counters, example_ossec_ids_alert_counters_2]
        aggregated_ossec_ids_alert_counters = thread.get_aggregated_ids_alert_counters()

        assert (aggregated_ossec_ids_alert_counters.severe_alerts ==
                example_ossec_ids_alert_counters.severe_alerts + example_ossec_ids_alert_counters_2.severe_alerts)
        assert (aggregated_ossec_ids_alert_counters.warning_alerts ==
                example_ossec_ids_alert_counters.warning_alerts + example_ossec_ids_alert_counters_2.warning_alerts)
        assert (aggregated_ossec_ids_alert_counters.total_alerts ==
                example_ossec_ids_alert_counters.total_alerts + example_ossec_ids_alert_counters_2.total_alerts)
        assert (round(aggregated_ossec_ids_alert_counters.alerts_weighted_by_level, 3) ==
                round(example_ossec_ids_alert_counters.alerts_weighted_by_level +
                      example_ossec_ids_alert_counters_2.alerts_weighted_by_level, 3))
        for idx in range(len(aggregated_ossec_ids_alert_counters.level_alerts)):
            assert (aggregated_ossec_ids_alert_counters.level_alerts[idx] ==
                    example_ossec_ids_alert_counters.level_alerts[idx] +
                    example_ossec_ids_alert_counters_2.level_alerts[idx])
        for idx in range(len(aggregated_ossec_ids_alert_counters.group_alerts)):
            assert (aggregated_ossec_ids_alert_counters.group_alerts[idx] ==
                    example_ossec_ids_alert_counters.group_alerts[idx] +
                    example_ossec_ids_alert_counters_2.group_alerts[idx])

    def test_aggregated_snort_ids_log_consumer_thread(self) -> None:
        """
        Tests creation of a snort ids log consumer thread and its methods

        :param example_snort_ids_alert_counters: an object of SnortIdsAlertCounters
        :return: None
        """

        example_snort_ids_alert_counters = SnortIdsAlertCounters()

        example_snort_ids_alert_counters.total_alerts = 10
        example_snort_ids_alert_counters.severe_alerts = 54
        example_snort_ids_alert_counters.warning_alerts = 3
        example_snort_ids_alert_counters.alerts_weighted_by_priority = 0.7
        example_snort_ids_alert_counters.ip = "1.2.3.4"
        example_snort_ids_alert_counters.ts = 1234

        example_snort_ids_alert_counters.priority_alerts = list(np.zeros(4))
        for idx in range(4):
            example_snort_ids_alert_counters.priority_alerts[idx] = random.randint(0, 10)

        example_snort_ids_alert_counters.class_alerts = list(
            np.zeros(len(set(constants.SNORT_IDS_ROUTER.SNORT_ALERT_IDS_ID.values()))))
        for idx in range(len(example_snort_ids_alert_counters.class_alerts)):
            example_snort_ids_alert_counters.class_alerts[idx] = random.randint(0, 10)

        thread = AggregatedSnortIdsLogConsumerThread(
            kafka_server_ip="1.2.3.4", kafka_port=1234, snort_ids_alert_counters=example_snort_ids_alert_counters.copy()
        )

        assert thread.snort_ids_alert_counters == example_snort_ids_alert_counters
        assert thread.kafka_port == 1234
        assert thread.kafka_server_ip == "1.2.3.4"

        # Test empty list
        thread.snort_ids_alert_counters_list = []
        aggregated_snort_ids_alert_counters = thread.get_aggregated_ids_alert_counters()
        assert aggregated_snort_ids_alert_counters == example_snort_ids_alert_counters

        # Test singleton list
        thread.snort_ids_alert_counters_list = [example_snort_ids_alert_counters.copy()]
        aggregated_snort_ids_alert_counters = thread.get_aggregated_ids_alert_counters()
        assert aggregated_snort_ids_alert_counters == example_snort_ids_alert_counters

        # Test list with 10 identical objects
        thread.snort_ids_alert_counters_list = []
        for i in range(10):
            thread.snort_ids_alert_counters_list.append(example_snort_ids_alert_counters.copy())
        aggregated_snort_ids_alert_counters = thread.get_aggregated_ids_alert_counters()
        assert (aggregated_snort_ids_alert_counters.total_alerts ==
                10 * example_snort_ids_alert_counters.total_alerts)
        assert (aggregated_snort_ids_alert_counters.severe_alerts ==
                10 * example_snort_ids_alert_counters.severe_alerts)
        assert (aggregated_snort_ids_alert_counters.warning_alerts ==
                10 * example_snort_ids_alert_counters.warning_alerts)
        assert (round(aggregated_snort_ids_alert_counters.alerts_weighted_by_priority, 3) ==
                round(10 * example_snort_ids_alert_counters.alerts_weighted_by_priority, 3))
        for idx in range(len(aggregated_snort_ids_alert_counters.priority_alerts)):
            assert (aggregated_snort_ids_alert_counters.priority_alerts[idx] ==
                    10 * example_snort_ids_alert_counters.priority_alerts[idx])
        for idx in range(len(aggregated_snort_ids_alert_counters.class_alerts)):
            assert (aggregated_snort_ids_alert_counters.class_alerts[idx] ==
                    10 * example_snort_ids_alert_counters.class_alerts[idx])

        # Test list with two different objects
        example_snort_ids_alert_counters_2 = SnortIdsAlertCounters()
        example_snort_ids_alert_counters_2.priority_alerts = list(np.zeros(4))
        for idx in range(4):
            example_snort_ids_alert_counters_2.priority_alerts[idx] = random.randint(0, 10)

        example_snort_ids_alert_counters_2.class_alerts = list(
            np.zeros(len(set(constants.SNORT_IDS_ROUTER.SNORT_ALERT_IDS_ID.values()))))
        for idx in range(len(example_snort_ids_alert_counters_2.class_alerts)):
            example_snort_ids_alert_counters_2.class_alerts[idx] = random.randint(0, 10)

        example_snort_ids_alert_counters_2.total_alerts = 15
        example_snort_ids_alert_counters_2.severe_alerts = 1
        example_snort_ids_alert_counters_2.warning_alerts = 18
        example_snort_ids_alert_counters_2.alerts_weighted_by_priority = 0.7

        thread.snort_ids_alert_counters_list = [example_snort_ids_alert_counters,
                                                example_snort_ids_alert_counters_2]
        aggregated_snort_ids_alert_counters = thread.get_aggregated_ids_alert_counters()

        assert (aggregated_snort_ids_alert_counters.severe_alerts ==
                example_snort_ids_alert_counters.severe_alerts + example_snort_ids_alert_counters_2.severe_alerts)
        assert (aggregated_snort_ids_alert_counters.warning_alerts ==
                example_snort_ids_alert_counters.warning_alerts + example_snort_ids_alert_counters_2.warning_alerts)
        assert (aggregated_snort_ids_alert_counters.total_alerts ==
                example_snort_ids_alert_counters.total_alerts + example_snort_ids_alert_counters_2.total_alerts)
        assert (round(aggregated_snort_ids_alert_counters.alerts_weighted_by_priority, 3) ==
                round(example_snort_ids_alert_counters.alerts_weighted_by_priority +
                      example_snort_ids_alert_counters_2.alerts_weighted_by_priority, 3))
        for idx in range(len(aggregated_snort_ids_alert_counters.priority_alerts)):
            assert (aggregated_snort_ids_alert_counters.priority_alerts[idx] ==
                    example_snort_ids_alert_counters.priority_alerts[idx] +
                    example_snort_ids_alert_counters_2.priority_alerts[idx])
        for idx in range(len(aggregated_snort_ids_alert_counters.class_alerts)):
            assert (aggregated_snort_ids_alert_counters.class_alerts[idx] ==
                    example_snort_ids_alert_counters.class_alerts[idx] +
                    example_snort_ids_alert_counters_2.class_alerts[idx])

    def test_aggregated_snort_ids_rule_log_consumer_thread(self) -> None:
        """
        Tests creation of a snort ids rule log consumer thread and its methods

        :param example_snort_ids_rule_alert_counters: an object of SnortIdsRuleAlertCounters
        :return: None
        """
        example_snort_ids_rule_alert_counters = SnortIdsRuleCounters()

        example_snort_ids_rule_alert_counters.rule_alerts = {}
        example_snort_ids_rule_alert_counters.rule_alerts[1] = 10
        example_snort_ids_rule_alert_counters.rule_alerts[2] = 3
        example_snort_ids_rule_alert_counters.ip = "1.2.3.4"
        example_snort_ids_rule_alert_counters.ts = 1234

        thread = AggregatedSnortIdsRuleLogConsumerThread(
            kafka_server_ip="1.2.3.4", kafka_port=1234,
            snort_ids_rule_counters=example_snort_ids_rule_alert_counters.copy()
        )

        assert thread.snort_ids_rule_counters == example_snort_ids_rule_alert_counters
        assert thread.kafka_port == 1234
        assert thread.kafka_server_ip == "1.2.3.4"

        # Test empty list
        thread.snort_ids_rule_counters_list = []
        aggregated_snort_ids_rule_alert_counters = thread.get_aggregated_ids_rule_counters()
        assert aggregated_snort_ids_rule_alert_counters == example_snort_ids_rule_alert_counters

        # Test singleton list
        thread.snort_ids_rule_counters_list = [example_snort_ids_rule_alert_counters.copy()]
        aggregated_snort_ids_rule_alert_counters = thread.get_aggregated_ids_rule_counters()
        assert aggregated_snort_ids_rule_alert_counters == example_snort_ids_rule_alert_counters

        # Test list with 10 identical objects
        thread.snort_ids_rule_counters_list = [example_snort_ids_rule_alert_counters.copy()] * 10
        aggregated_snort_ids_rule_alert_counters = thread.get_aggregated_ids_rule_counters()

        for k, v in aggregated_snort_ids_rule_alert_counters.rule_alerts.items():
            assert (aggregated_snort_ids_rule_alert_counters.rule_alerts[k] ==
                    10 * example_snort_ids_rule_alert_counters.rule_alerts[k])

        # Test list with two different objects
        example_snort_ids_rule_alert_counters_2 = SnortIdsRuleCounters()
        example_snort_ids_rule_alert_counters_2.rule_alerts = {}
        example_snort_ids_rule_alert_counters_2.rule_alerts[1] = 12
        example_snort_ids_rule_alert_counters_2.rule_alerts[2] = 8
        example_snort_ids_rule_alert_counters_2.ip = "1.2.3.4"
        example_snort_ids_rule_alert_counters_2.ts = 1234

        thread.snort_ids_rule_counters_list = [example_snort_ids_rule_alert_counters,
                                               example_snort_ids_rule_alert_counters_2]
        aggregated_snort_ids_rule_alert_counters = thread.get_aggregated_ids_rule_counters()

        for k, v in aggregated_snort_ids_rule_alert_counters.rule_alerts.items():
            assert (aggregated_snort_ids_rule_alert_counters.rule_alerts[k] ==
                    example_snort_ids_rule_alert_counters.rule_alerts[k] +
                    example_snort_ids_rule_alert_counters_2.rule_alerts[k])

    def test_attacker_actions_consumer_thread(self) -> None:
        """
        Tests creation of a attacker action consumer thread and its methods

        :return: None
        """
        example_attacker_actions_consumer = EmulationAttackerAction(
            id=EmulationAttackerActionId.HTTP_ENUM_HOST, name="test1", cmds=["test1"],
            type=EmulationAttackerActionType.CONTINUE, descr="test", ips=["1.1.2.3"], index=1)
        thread = AttackerActionsConsumerThread(kafka_server_ip="1.1.1.1", kafka_port=1234,
                                               attacker_actions=[example_attacker_actions_consumer])
        assert thread.attacker_actions[0] == example_attacker_actions_consumer

    def test_avg_host_metrics_thread(self) -> None:
        """
        Tests creation of a avg host metrics thread and its methods

        :return: None
        """
        example_host_metrics = HostMetrics(num_logged_in_users=1, num_failed_login_attempts=10,
                                           num_open_connections=1, num_login_events=2,
                                           num_processes=5, num_users=1, ip="1.1.1.1", ts=0.5)
        emulation_defender_machine_observation_state_example = EmulationDefenderMachineObservationState(
            ips=["test"], kafka_config=None, host_metrics=example_host_metrics,
            docker_stats=DockerStats(), snort_ids_ip_alert_counters=SnortIdsIPAlertCounters(),
            ossec_ids_alert_counters=OSSECIdsAlertCounters())
        thread = AvgHostMetricsThread(host_metrics=example_host_metrics,
                                      machines=[emulation_defender_machine_observation_state_example], sleep_time=1)
        assert thread.host_metrics == example_host_metrics
        assert thread.machines[0] == emulation_defender_machine_observation_state_example

    def test_client_population_consumer_thread(self) -> None:
        """
        Tests creation of a client population consumer thread and its methods

        :return: None
        """
        example_client_population_metric = ClientPopulationMetrics()
        thread = ClientPopulationConsumerThread(kafka_server_ip="1.1.1.1", kafka_port=1234,
                                                client_population_metrics=example_client_population_metric)
        assert thread.client_population_metrics == example_client_population_metric
        assert thread.kafka_server_ip == "1.1.1.1"
        assert thread.kafka_port == 1234

        # Test empty list
        thread.client_population_metrics_list = []
        avg_client_population_metrics = thread.get_average_client_population_metrics()
        assert avg_client_population_metrics == example_client_population_metric

        # Test singleton list
        thread.client_population_metrics_list = [example_client_population_metric.copy()]
        avg_client_population_metrics = thread.get_average_client_population_metrics()
        assert avg_client_population_metrics == example_client_population_metric

        # Test list with 10 identical objects
        thread.client_population_metrics_list = []
        for i in range(10):
            thread.client_population_metrics_list.append(example_client_population_metric.copy())
        avg_client_population_metrics = thread.get_average_client_population_metrics()

        assert avg_client_population_metrics.num_clients == example_client_population_metric.num_clients
        assert avg_client_population_metrics.rate == example_client_population_metric.rate

        example_client_population_metric_1 = ClientPopulationMetrics(num_clients=10, rate=40)
        thread.client_population_metrics_list = [example_client_population_metric.copy(),
                                                 example_client_population_metric_1]

        avg_client_population_metrics = thread.get_average_client_population_metrics()

        assert (avg_client_population_metrics.num_clients ==
                round((example_client_population_metric.num_clients +
                       example_client_population_metric_1.num_clients) / 2))
        assert (avg_client_population_metrics.rate ==
                round(example_client_population_metric.rate + example_client_population_metric_1.rate) / 2)

    def test_defender_actions_consumer_thread(self) -> None:
        """
        Tests creation of a defender actions consumer thread and its methods

        :return: None
        """
        example_defender_action = EmulationDefenderAction(id=EmulationDefenderActionId.CONTINUE, name="test",
                                                          cmds=["test"], type=EmulationDefenderActionType.CONTINUE,
                                                          descr="test", ips=["1.1.1.1"], index=1)
        thread = DefenderActionsConsumerThread(kafka_server_ip="1.1.1.1", kafka_port=1234,
                                               defender_actions=[example_defender_action])

        assert thread.kafka_server_ip == "1.1.1.1"
        assert thread.kafka_port == 1234
        assert thread.defender_actions[0] == example_defender_action

    def test_docker_host_stats_consumer_thread(self) -> None:
        """
        Tests creation of a docker host stats consumer thread and its methods

        :return: None
        """
        example_docker_host_stats = DockerStats()
        thread = DockerHostStatsConsumerThread(kafka_server_ip="1.1.1.1", kafka_port=1234,
                                               docker_stats=example_docker_host_stats, host_ip="1.2.3.4")

        assert thread.kafka_server_ip == "1.1.1.1"
        assert thread.kafka_port == 1234
        assert thread.docker_stats == example_docker_host_stats
        assert thread.host_ip == "1.2.3.4"

    def test_docker_stats_consumer_thread(self) -> None:
        """
        Tests creation of a docker stats consumer thread and its methods

        :return: None
        """
        example_docker_host_stats = DockerStats()
        thread = DockerStatsConsumerThread(kafka_server_ip="1.1.1.1", kafka_port=1234,
                                           docker_stats=example_docker_host_stats)

        assert thread.kafka_server_ip == "1.1.1.1"
        assert thread.kafka_port == 1234
        assert thread.docker_stats == example_docker_host_stats

    def test_ossec_ids_log_consumer_thread(self) -> None:
        """
        Tests creation of a ossec ids log consumer thread and its methods

        :return: None
        """
        example_ossec_ids_alerts = OSSECIdsAlertCounters()
        thread = OSSECIdsLogConsumerThread(host_ip="1.2.3.4", kafka_server_ip="1.1.1.1", kafka_port=1234,
                                           ossec_ids_alert_counters=example_ossec_ids_alerts)

        assert thread.kafka_server_ip == "1.1.1.1"
        assert thread.kafka_port == 1234
        assert thread.ossec_ids_alert_counters == example_ossec_ids_alerts
        assert thread.host_ip == "1.2.3.4"

    def test_snort_ids_log_consumer_thread(self) -> None:
        """
        Tests creation of a snort ids log consumer thread and its methods

        :return: None
        """
        example_snort_ids_alerts_counter = SnortIdsIPAlertCounters()
        thread = SnortIdsLogConsumerThread(host_ip="1.2.3.4", kafka_server_ip="1.1.1.1", kafka_port=1234,
                                           snort_ids_alert_counters=example_snort_ids_alerts_counter)

        assert thread.kafka_server_ip == "1.1.1.1"
        assert thread.kafka_port == 1234
        assert thread.snort_ids_alert_counters == example_snort_ids_alerts_counter
        assert thread.host_ip == "1.2.3.4"



from typing import List, Union, Dict, Any
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class ContainersConfig(JSONSerializable):
    """
    A DTO representing the configuration of the containers that make up an emulation environment
    """

    def __init__(self, containers: List[NodeContainerConfig], agent_ip: str, router_ip: str,
                 networks: List[ContainerNetwork], ids_enabled: bool,
                 vulnerable_nodes=None, agent_reachable_nodes=None):
        """
        Initializes the DTO

        :param containers: the list of containers
        :param agent_ip: the ip of the agent
        :param router_ip: the ip of the router
        :param ids_enabled: whether the IDS is enabled or nt
        :param vulnerable_nodes: the list of vulnerable nodes
        :param networks: list of subnetworks
        :param agent_reachable_nodes: nodes directly reachable by the attacker
        """
        self.containers = containers
        self.agent_ip = agent_ip
        self.router_ip = router_ip
        self.ids_enabled = ids_enabled
        self.vulnerable_nodes = vulnerable_nodes
        self.networks = networks
        self.agent_reachable_nodes = agent_reachable_nodes

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ContainersConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ContainersConfig(
            containers=list(map(lambda x: NodeContainerConfig.from_dict(x), d["containers"])),
            agent_ip=d["agent_ip"], router_ip=d["router_ip"],
            networks=list(map(lambda x: ContainerNetwork.from_dict(x), d["networks"])),
            ids_enabled=d["ids_enabled"], vulnerable_nodes=d["vulnerable_nodes"],
            agent_reachable_nodes=d["agent_reachable_nodes"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["agent_ip"] = self.agent_ip
        d["router_ip"] = self.router_ip
        d["networks"] = list(map(lambda x: x.to_dict(), self.networks))
        d["ids_enabled"] = self.ids_enabled
        d["vulnerable_nodes"] = self.vulnerable_nodes
        d["containers"] = list(map(lambda x: x.to_dict(), self.containers))
        d["agent_reachable_nodes"] = self.agent_reachable_nodes
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"containers:{self.containers},networks:{self.networks},agent_ip:{self.agent_ip}, " \
               f"router_ip:{self.router_ip}" \
               f"ids_enabled:{self.ids_enabled},vulnerable_nodes:{self.vulnerable_nodes}, " \
               f"agent_reachable_nodes: {self.agent_reachable_nodes}"

    def get_reachable_ips(self, container: NodeContainerConfig) -> List[str]:
        """
        Get list of IP addresses reachable from a given container

        :param container: the container to get reachable IPs from
        :return:
        """
        reachable_ips = []
        for c in self.containers:
            for ip_net in c.ips_and_networks:
                ip, net = ip_net
                for container_ip_net in container.ips_and_networks:
                    container_ip, container_net = container_ip_net
                    if net.name == container_net.name:
                        reachable_ips.append(ip)
        return reachable_ips

    def get_agent_container(self) -> Union[NodeContainerConfig, None]:
        """
        :return: get container of the attacker agent
        """
        for container in self.containers:
            if self.agent_ip in container.get_ips():
                return container
        return None

    def get_agent_reachable_ips(self) -> List[str]:
        """
        :return: list of ips reachable for the attacker agent
        """
        agent_container = self.get_agent_container()
        if agent_container is None:
            raise ValueError("Agent container is None")
        return self.get_reachable_ips(container=agent_container)

    def get_container_from_ip(self, ip: str) -> Union[NodeContainerConfig, None]:
        """
        Utility function for getting the container

        :param ip: the ip of the container
        :return: the container with the given ip or None
        """
        for c in self.containers:
            if ip in c.get_ips() or ip == c.docker_gw_bridge_ip:
                return c
        return None

    def get_container_from_full_name(self, name: str) -> Union[NodeContainerConfig, None]:
        """
        Utility function for getting the container

        :param name: the full name of the container
        :return: the container with the given ip or None
        """
        for c in self.containers:
            if name == c.full_name_str:
                return c
        return None

    @staticmethod
    def from_json_file(json_file_path: str) -> "ContainersConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ContainersConfig.from_dict(json.loads(json_str))

    def copy(self) -> "ContainersConfig":
        """
        :return: a copy of the DTO
        """
        return ContainersConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int, physical_servers: List[str]) -> "ContainersConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :param physical_servers: the physical servers where the execution will be deployed
        :return: the new config
        """
        config = self.copy()
        config.containers = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet,
                                                                         physical_servers=physical_servers),
                                     config.containers))
        config.networks = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                   config.networks))
        config.agent_ip = GeneralUtil.replace_first_octet_of_ip(ip=config.agent_ip, ip_first_octet=ip_first_octet)
        config.router_ip = GeneralUtil.replace_first_octet_of_ip(ip=config.router_ip, ip_first_octet=ip_first_octet)
        config.vulnerable_nodes = list(map(lambda x: GeneralUtil.replace_first_octet_of_ip(
            ip=x, ip_first_octet=ip_first_octet), config.vulnerable_nodes))
        config.agent_reachable_nodes = list(map(lambda x: GeneralUtil.replace_first_octet_of_ip(
            ip=x, ip_first_octet=ip_first_octet), config.agent_reachable_nodes))
        return config



from typing import Optional, List
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome


class EmulationAttackerNetworkServiceActions:
    """
    Class that implements network-service actions for the attacker in the emulation
    """

    @staticmethod
    def SERVICE_LOGIN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Action for using known credentials to login to a server

        :param index: index of the machine to apply the action to
        :param ips: ip of the machine to apply the action to
        :return: the action
        """
        if ips is None:
            ips = []
        id = EmulationAttackerActionId.NETWORK_SERVICE_LOGIN
        return EmulationAttackerAction(id=id, name="Network service login", cmds=[],
                                       type=EmulationAttackerActionType.POST_EXPLOIT,
                                       descr="Uses known credentials to login to network services on a server",
                                       index=index,
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.LOGIN)
